/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <SDL_opengl.h>
#include "box.h"
#include "vect.h"
#include "matrix.h"

class Quad {
private:
	GLfloat color[4];
	GLfloat vertex1[3], vertex2[3], vertex3[3], vertex4[3];
	GLfloat normal[3];
public:
	GLfloat z;
	Quad() {
	}
	Quad(const GLfloat *c, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3, const GLfloat *v4, const GLfloat *n, const GLfloat *pm) {
		memcpy(color,c,sizeof(GLfloat)*4);
		memcpy(vertex1,v1,sizeof(GLfloat)*3);
		memcpy(vertex2,v2,sizeof(GLfloat)*3);
		memcpy(vertex3,v3,sizeof(GLfloat)*3);
		memcpy(vertex4,v4,sizeof(GLfloat)*3);
		memcpy(normal,n,sizeof(GLfloat)*3);
		if (pm!=NULL) {
			GLfloat tmp[3];
			Vect_copy3f(tmp,v1);
			Vect_add3f(tmp,v2);
			Vect_add3f(tmp,v3);
			Vect_add3f(tmp,v4);
			Vect_mult3f(tmp,0.25);
			GLfloat proj[3];
			Matrix_multVect3f(proj,pm,tmp,1);
			z = proj[2];
		}
	}
	void draw() {
		glColor4fv(color);
		glNormal3fv(normal);
		glVertex3fv(vertex1);
		glVertex3fv(vertex2);
		glVertex3fv(vertex3);
		glVertex3fv(vertex4);
	}
};

int Quad_comp(const void *p1, const void *p2) {
	Quad *q1 = (Quad*)p1;
	Quad *q2 = (Quad*)p2;
	if (q1->z<q2->z) return -1;
	if (q1->z>q2->z) return 1;
	return 0;
}


inline void drawVertex(const GLfloat *pos, const GLfloat *size, int x, int y, int z) {
	GLfloat vertex[3];
	vertex[0] = pos[0] + size[0]*x;
	vertex[1] = pos[1] + size[1]*y;
	vertex[2] = pos[2] + size[2]*z;
	glVertex3fv(vertex);
}
inline void setVertex(GLfloat *vertex, const GLfloat *pos, const GLfloat *size, int x, int y, int z) {
	vertex[0] = pos[0] + size[0]*x;
	vertex[1] = pos[1] + size[1]*y;
	vertex[2] = pos[2] + size[2]*z;
}


void getBox(const Box &b, Quad *quads, GLfloat *PM) {
	GLfloat normal[3];
	GLfloat vertex1[3], vertex2[3], vertex3[3], vertex4[3];
#define SETNORMAL(x,y,z) normal[0] = x; normal[1] = y; normal[2] = z;

	SETNORMAL(0,0,b.size[2]>0?-1:1);
	setVertex(vertex1,b.pos,b.size,0,0,0);
	setVertex(vertex2,b.pos,b.size,0,1,0);
	setVertex(vertex3,b.pos,b.size,1,1,0);
	setVertex(vertex4,b.pos,b.size,1,0,0);
	quads[0] = Quad(b.color, vertex1, vertex2, vertex3, vertex4, normal, PM);

	SETNORMAL(0,0,b.size[2]>0?1:-1);
	setVertex(vertex1,b.pos,b.size,0,0,1);
	setVertex(vertex2,b.pos,b.size,1,0,1);
	setVertex(vertex3,b.pos,b.size,1,1,1);
	setVertex(vertex4,b.pos,b.size,0,1,1);
	quads[1] = Quad(b.color, vertex1, vertex2, vertex3, vertex4, normal, PM);

	SETNORMAL(0,b.size[1]>0?-1:1,0);
	setVertex(vertex1,b.pos,b.size,0,0,0);
	setVertex(vertex2,b.pos,b.size,1,0,0);
	setVertex(vertex3,b.pos,b.size,1,0,1);
	setVertex(vertex4,b.pos,b.size,0,0,1);
	quads[2] = Quad(b.color, vertex1, vertex2, vertex3, vertex4, normal, PM);

	SETNORMAL(0,b.size[1]>0?1:-1,0);
	setVertex(vertex1,b.pos,b.size,0,1,0);
	setVertex(vertex2,b.pos,b.size,0,1,1);
	setVertex(vertex3,b.pos,b.size,1,1,1);
	setVertex(vertex4,b.pos,b.size,1,1,0);
	quads[3] = Quad(b.color, vertex1, vertex2, vertex3, vertex4, normal, PM);

	SETNORMAL(b.size[0]>0?-1:1,0,0);
	setVertex(vertex1,b.pos,b.size,0,0,0);
	setVertex(vertex2,b.pos,b.size,0,0,1);
	setVertex(vertex3,b.pos,b.size,0,1,1);
	setVertex(vertex4,b.pos,b.size,0,1,0);
	quads[4] = Quad(b.color, vertex1, vertex2, vertex3, vertex4, normal, PM);

	SETNORMAL(b.size[0]>0?1:-1,0,0);
	setVertex(vertex1,b.pos,b.size,1,0,0);
	setVertex(vertex2,b.pos,b.size,1,1,0);
	setVertex(vertex3,b.pos,b.size,1,1,1);
	setVertex(vertex4,b.pos,b.size,1,0,1);
	quads[5] = Quad(b.color, vertex1, vertex2, vertex3, vertex4, normal, PM);

}

void drawBox(const Box &b, bool begin) {
	if (begin) {
		int neg = 0;
		if (b.size[0]<0) neg++;
		if (b.size[1]<0) neg++;
		if (b.size[2]<0) neg++;
		if (neg%2==0)
			glFrontFace(GL_CCW);
		else
			glFrontFace(GL_CW);
		glBegin(GL_QUADS);
	}
	Quad quads[6];
	getBox(b,quads,NULL);
	for (int i = 0; i < 6; i++)
		quads[i].draw();

	if (begin)
		glEnd();
}


void drawBox(GLfloat c0, GLfloat c1, GLfloat c2, GLfloat p0, GLfloat p1, GLfloat p2, GLfloat s0, GLfloat s1, GLfloat s2, bool begin) {
	Box b(c0,c1,c2,1, p0,p1,p2, s0,s1,s2);
	drawBox(b,begin);
}


void drawBoxes(const vector<Box> &boxes) {
	int n = boxes.size();
	vector<int> alpha;
	glDisable(GL_BLEND);
	glEnable(GL_CULL_FACE);
	for (int i = 0; i < n; i++) {
		const Box &b = boxes[i];
		glColor4fv(b.color);
		if (b.color[3] < 0.99) {
			alpha.push_back(i);
			continue;
		}
		drawBox(b,true);
	}
	glFrontFace(GL_CCW);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	n = alpha.size();
	GLfloat modelview[16];
	GLfloat projection[16];
	GLfloat PM[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, modelview);
	glGetFloatv(GL_PROJECTION_MATRIX, projection);
	Matrix_multf(PM,projection,modelview);
	Quad *quads = new Quad[n*6];
	for (int i = 0; i < n; i++) {
		const Box &b = boxes[alpha[i]];
		getBox(b,&quads[i*6],PM);
	}
	n *= 6;
	qsort(quads,n,sizeof(Quad),Quad_comp);
	glBegin(GL_QUADS);
	for (int i = n-1; i >= 0; i--) {
		quads[i].draw();
	}
	glEnd();
	delete [] quads;
	glDisable(GL_BLEND);
}



