/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <SDL_opengl.h>
#include "camera.h"
#include "vect.h"
#include "vect.h"
#include "matrix.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

static void getCameraPos(const GLfloat *upleft, const GLfloat *upright, const GLfloat *downleft, GLint *viewport, float fovy, bool project, Vect3 &width, Vect3 &height, Vect3 &middle, Vect3 &dir, float &fovx, float &dist) {
	Vect3 ul(upleft);
	Vect3 ur(upright);
	Vect3 dl(downleft);
	if (project) {
		GLfloat modelview[16];
		glGetFloatv(GL_MODELVIEW_MATRIX,modelview);
		GLfloat tmp[3];
		Matrix_multVect3f(tmp,modelview,upleft,1);
		ul = Vect3(tmp);
		Matrix_multVect3f(tmp,modelview,upright,1);
		ur = Vect3(tmp);
		Matrix_multVect3f(tmp,modelview,downleft,1);
		dl = Vect3(tmp);
	}
	fovy *= M_PI/180;
	float aspect = float(viewport[2])/viewport[3];
	height = dl - ul;
	width = ur - ul;
	middle = ul + 0.5*width + 0.5*height;
	dir = Vect3::cross(height,width).normalize();
	float dist1 = (height.abs()/2)/tan(fovy/2);
	fovx = 2*atan(aspect*tan(fovy/2));
	float dist2 = (width.abs()/2)/tan(fovx/2);
	dist = dist1>dist2?dist1:dist2;
}

void getCameraLimits(GLfloat *upleft, GLfloat *upright, GLfloat *downleft, GLint *viewport, float fovy, bool project, float distdiff) {
	Vect3 width, height, middle, dir;
	float fovx, dist;
	getCameraPos(upleft,upright,downleft,viewport,fovy,project,width,height,middle,dir,fovx,dist);
	dist += distdiff;
	float h = dist*tan(fovy/2);
	float w = dist*tan(fovx/2);
	width.normalize();
	height.normalize();
	Vect3 ul = middle - w*width - h*height - distdiff*dir;
	Vect3 ur = middle + w*width - h*height - distdiff*dir;
	Vect3 dl = middle - w*width + h*height - distdiff*dir;
	ul.get(upleft);
	ur.get(upright);
	dl.get(downleft);
}

void setCamera(const GLfloat *upleft, const GLfloat *upright, const GLfloat *downleft, float fovy, bool project, GLfloat *matrix, GLfloat *campos) {
	Vect3 width, height, middle, dir;
	float fovx, dist;
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT,viewport);
	getCameraPos(upleft,upright,downleft,viewport,fovy,project,width,height,middle,dir,fovx,dist);
	Vect3 pos = middle + dir*dist;
	height *= -1;
	GLfloat up[3], eye[3], center[3];
	height.get(up);
	pos.get(eye);
	if (campos) pos.get(campos);
	middle.get(center);
	if (matrix!=NULL) {
		glPushMatrix();
		glLoadIdentity();
		gluLookAt(eye[0],eye[1],eye[2],center[0],center[1],center[2],up[0],up[1],up[2]);
		GLint mode;
		glGetIntegerv(GL_MATRIX_MODE,&mode);
		if (mode==GL_MODELVIEW)
			glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
		else if (mode==GL_PROJECTION)
			glGetFloatv(GL_PROJECTION_MATRIX,matrix);
		glPopMatrix();
	} else {
		gluLookAt(eye[0],eye[1],eye[2],center[0],center[1],center[2],up[0],up[1],up[2]);
	}
}

