/*
    Cothread
    Copyright 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include "cothread.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#ifdef COTHREAD_ASM_SUPPORTED

#define STACKSIZE 8192
#define STACKFILL 0x42
/* #define STACKTEST */

struct context {
	unsigned int data[2]; /* x86: esp, ebp, ppc: r1, r31 */
};

struct _cothread {
	void *stack;

	int done;
	int join;
	int result;

	int started;
	cothread_func func;
	void *arg;

	int current;
	struct context ctx[2];
};

#if defined(WIN32) || defined(__APPLE__)
#define FUNCNAME "_thread_init"
#else
#define FUNCNAME "thread_init"
#endif

static void __attribute__((used)) thread_init(cothread thread) {
	thread->started = 1;
	thread->result = thread->func(thread->arg);

	/* don't ever exit */
	thread->done = 1;
	cothread_swap(thread);
}

cothread cothread_create(cothread_func func, void* arg) {
	cothread thread = (cothread) malloc(sizeof(struct _cothread));

	thread->done = 0;
	thread->join = 0;
	thread->started = 0;
	thread->func = func;
	thread->arg = arg;
	thread->current = 0;
	thread->stack = malloc(STACKSIZE);
#ifdef STACKTEST
	memset(thread->stack, STACKFILL, STACKSIZE);
#endif
	thread->ctx[1].data[0] = (unsigned int) ((unsigned char*)thread->stack + STACKSIZE);
#if defined(__ppc__) || defined(__powerpc__)
	/* align to 16 byte border */
	thread->ctx[1].data[0] = (thread->ctx[1].data[0] - 28)&0xFFFFFFF0;
	/* first back chain word must be zero */
	*((unsigned int*)thread->ctx[1].data[0]) = 0;
#endif
	thread->ctx[1].data[1] = (unsigned int) 0;

	return thread;
}

void __attribute__((noinline)) cothread_swap(cothread thread) {
	struct context* curctx;
	struct context* newctx;

	if (thread->current == 0 && thread->done) return;
	if (thread->current == 1 && thread->join && !thread->done) return;

	curctx = &thread->ctx[thread->current];
	thread->current = (thread->current+1)&1;
	newctx = &thread->ctx[thread->current];

#ifdef __i386__
	__asm__ volatile(
		"pushl %%ebx\n\t"
		"movl %%esp, 0(%0)\n\t"
		"movl %%ebp, 4(%0)\n\t"

		"cmpl $0, %3\n\t"
		"jnz 1f\n\t"

		"movl 0(%1), %%esp\n\t"
		"xorl %%ebp, %%ebp\n\t"
		"pushl %2\n\t"
		"call " FUNCNAME "\n"

		"1:\n\t"
		"movl 0(%1), %%esp\n\t"
		"movl 4(%1), %%ebp\n\t"
		"popl %%ebx\n\t"
		: /*output*/
		: /*input*/ "S" (curctx->data), "D" (newctx->data), "r" (thread), "g" (thread->started)
		: /*clobber*/ "memory");
#elif defined(__ppc__) || defined(__powerpc__)
	__asm__ volatile(
#ifdef __APPLE__
		"stw r31, 4(%0)\n\t"
		"stw r1, 0(%0)\n\t"
 
		"cmpwi cr7, %3, 0\n\t"
		"bne cr7, L%=\n\t"
 
		"lwz r1, 0(%1)\n\t"
		"xor r0, r0, r0\n\t"
		"mtlr r0\n\t"
		"mr r3, %2\n\t"
		"bl " FUNCNAME "\n"

		"L%=:\n\t"
		"lwz r1, 0(%1)\n\t"
		"lwz r31, 4(%1)\n\t"
#else
		"stw 31, 4(%0)\n\t"
		"stw 1, 0(%0)\n\t"
 
		"cmpwi cr7, %3, 0\n\t"
		"bne cr7, L%=\n\t"
 
		"lwz 1, 0(%1)\n\t"
		"xor 0, 0, 0\n\t"
		"mtlr 0\n\t"
		"mr 3, %2\n\t"
		"bl " FUNCNAME "\n"

		"L%=:\n\t"
		"lwz 1, 0(%1)\n\t"
		"lwz 31, 4(%1)\n\t"
#endif
		: /*output*/
		: /*input*/ "r" (curctx->data), "r" (newctx->data), "r" (thread), "r" (thread->started)
		: /*clobber*/ "r0", "r3", "cr7", "memory",
		"r2", "r13", "r14", "r15", "r16", "r17", "r18", "r19",
		"r20", "r21", "r22", "r23", "r24", "r25", "r26",
		"r27", "r28", "r29", "r30", /*"r31",*/
		"fr14", "fr15", "fr16", "fr17", "fr18", "fr19", "fr20",
		"fr21", "fr22", "fr23", "fr24", "fr25", "fr26", "fr27",
		"fr28", "fr29", "fr30", "fr31",
		"lr", "ctr", "xer", "cr2", "cr3", "cr4");
#endif
}

int cothread_join(cothread thread) {
	int result;

	thread->join = 1;
	cothread_swap(thread);

#ifdef STACKTEST
	{
		int n;
		for (n = 0; n < STACKSIZE; n++)
			if (((unsigned char*)thread->stack)[n] != STACKFILL)
				break;
		printf("stack used: %d\n", STACKSIZE-n);
	}
#endif
	free(thread->stack);
	result = thread->result;
	free(thread);
	return result;
}

int cothread_running(cothread thread) {
	return !thread->done;
}

int cothread_active(cothread thread) {
	return thread->current;
}

#endif

