/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <SDL_opengl.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <vector>
#include "fps.h"
#include "graph.h"
#include "graphanim.h"
#include "treeanim.h"
#include "tree.h"
#include "algorithms.h"
#include "sdlttf.h"

#if defined(__APPLE__) && defined(__MACH__)
#include <CoreFoundation/CoreFoundation.h>
#endif

using std::vector;



bool fullscreen = false;

bool done = false;
Font* treefont = NULL;
Font* namefont = NULL;




DataOrder order = RANDOM;
int treeelements = 25;
int sortelements = 60;



static void initFont() {
	const char *file = "VeraBd.ttf";
#if defined(__APPLE__) && defined(__MACH__)
	char buffer[1000];
	CFURLRef url = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle());
	CFURLGetFileSystemRepresentation(url, true, (UInt8*) buffer, sizeof(buffer));
	CFRelease(url);
	strcat(buffer, "/");
	strcat(buffer, file);
	file = buffer;
#endif
	treefont = new Font(file, 40, '0', '9');
	namefont = new Font(file, 20, 32, 127);
}

static void deleteFont() {
	delete treefont;
	delete namefont;
	treefont = namefont = NULL;
}

void resize(int w, int h) {
	if (h==0) h = 1;
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
//	glLoadIdentity();

}

void init() {
	glShadeModel(GL_SMOOTH);
	glClearColor(0,0,0,0);
	glClearDepth(1);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


//	glEnable(GL_NORMALIZE);
	glDisable(GL_TEXTURE_2D);

	glTranslatef(0,0,-10);

	deleteFont();
	initFont();

}


void drawBorder() {
	GLfloat viewport[4];
	glGetFloatv(GL_VIEWPORT, viewport);
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,viewport[2],0,viewport[3],-1,1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glColor4f(1,1,1,1);
	glDisable(GL_LIGHTING);

	glBegin(GL_LINE_LOOP);
	glVertex2f(0,0);
	glVertex2f(0,viewport[3]-1);
	glVertex2f(viewport[2]-1,viewport[3]-1);
	glVertex2f(viewport[2]-1,0);
	glEnd();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glPopAttrib();
}

FPSCounter counter(500);
int flags;
SDL_Surface *screen;
int starttime;
int lasttime;
bool active;
bool turbo;
bool slowmotion;
bool paused;

void printName(const char *str, int height) {

	int maxlen = 20;
	int y = height-7;
	int textHeight = 10;
	glColor4f(1, 1, 1, 1);
	int len = strlen(str);
	if (len > maxlen) {
		char buffer[1024];
		strcpy(buffer, str);
		str = buffer;
		int start = 0;
		int cur = maxlen;
		while (cur < len) {
			while (cur >= start && buffer[cur] != ' ')
				cur--;
			if (cur == start) {
				while (cur < len) {
					if (buffer[cur] == ' ')
						break;
					cur++;
				}
			}
			if (cur < len) {
				buffer[cur] = '\0';
				namefont->printOverlay2D(textHeight, 7, y, "%s", str);
				y -= textHeight+2;

				start = cur+1;
				cur = start + maxlen;
				str = &buffer[start];
			}
		}
	}
	namefont->printOverlay2D(textHeight, 7, y, "%s", str);
}

void drawFrame(Anim **anims) {
	SDL_Event event;
	while (SDL_PollEvent(&event)) {
		if (event.type==SDL_VIDEORESIZE) {
			char name[50];
			SDL_VideoDriverName(name,sizeof(name));
			if (!strcmp(name,"x11") || !strcmp(name,"Quartz")) {
				screen = SDL_SetVideoMode(event.resize.w,event.resize.h,screen->format->BitsPerPixel,flags);
				if (screen==NULL) {
					fprintf(stderr,"%s\n",SDL_GetError());
					exit(1);
				}
				if (!strcmp(name,"Quartz")) {
					init();
					reinitQuadrics();
				}
			} else {
				screen->w = event.resize.w;
				screen->h = event.resize.h;
			}
			resize(screen->w,screen->h);
		} else if (event.type==SDL_KEYDOWN) {
			switch (event.key.keysym.sym) {
			case SDLK_ESCAPE:
				done = true;
				break;
			case SDLK_F1:
				GLfloat matrix[16];
				glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
				fullscreen = !fullscreen;
				if (fullscreen) flags |= SDL_FULLSCREEN;
				else flags &= ~SDL_FULLSCREEN;
				screen = SDL_SetVideoMode(screen->w,screen->h,screen->format->BitsPerPixel,flags);
				init();
				reinitQuadrics();
				resize(screen->w,screen->h);
				if (screen==NULL) {
					fprintf(stderr,"%s\n",SDL_GetError());
					exit(1);
				}
				glMatrixMode(GL_MODELVIEW);
				glLoadMatrixf(matrix);
				break;
			case SDLK_RIGHT:
				turbo = true;
				break;
			case SDLK_LEFT:
				slowmotion = true;
				break;
			case SDLK_SPACE:
				paused = !paused;
				break;
			default:
				break;
			}
		} else if (event.type==SDL_KEYUP) {
			switch (event.key.keysym.sym) {
			case SDLK_RIGHT:
				turbo = false;
				break;
			case SDLK_LEFT:
				slowmotion = false;
				break;
			default:
				break;
			}
		} else if (event.type==SDL_QUIT) {
			done = true;
		} else if (event.type==SDL_ACTIVEEVENT) {
			if (event.active.state==SDL_APPACTIVE)
				active = event.active.gain;
		} else if (event.type==SDL_MOUSEBUTTONDOWN) {
		} else if (event.type==SDL_MOUSEBUTTONUP) {
		} else if (event.type==SDL_MOUSEMOTION) {
		}
	}
	if (done) return;

	int time = SDL_GetTicks()-starttime;
	int difftime = time-lasttime;
	lasttime = time;
	int n = 0;
	int addtime = difftime;
	if (turbo) addtime *= 10;
	if (slowmotion) {
		addtime /= 3;
		if (addtime == 0) addtime = 1;
	}
	if (paused) addtime = 0;
	for (int i = 0; anims[i]; i++) {
		anims[i]->addTime(addtime);
		n++;
	}
	if (!active) {
		SDL_Delay(15);
		return;
	}
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	const GLint &x = viewport[0], &y = viewport[1], &w = viewport[2], &h = viewport[3];
	for (int i = 0; i < n; i++) {
		int rows = 1, cols = 1;
/*		if (n==1) {
			glViewport(x,y,w,h);
		} else if (n==2) {
			if (i==0) glViewport(x,y,w/2,h);
			else glViewport(x+w/2,y,w-w/2,h);
		} else if (n==3) {
			if (i==0) glViewport(x,y,w/2,h);
			else if (i==1) glViewport(x+w/2,y+h/2,w-w/2,h-h/2);
			else glViewport(x+w/2,y,w-w/2,h/2);
		} else if (n==4) {
			if (i==0) glViewport(x,y+h/2,w/2,h-h/2);
			else if (i==1) glViewport(x+w/2,y+h/2,w-w/2,h-h/2);
			else if (i==2) glViewport(x,y,w/2,h/2);
			else glViewport(x+w/2,y,w-w/2,h/2);
		}
*/
		if (n <= 2) {
			rows = 1;
		} else if (n <= 6) {
			rows = 2;
		} else if (n <= 12) {
			rows = 3;
		} else if (n <= 20) {
			rows = 4;
		}
		cols = (n+rows-1)/rows;
		int winx = i/rows;
		int winy = (rows-1-i%rows);
		int winw = w/cols;
		int winh = h/rows;
		if (i==n-1 && winy>0) {
			winh *= 1+winy;
			winy = 0;
		}
		winx *= winw;
		winy *= winh;
		if (winy==rows-1)
			winh = h - winy*winh;
		if (winx==cols-1)
			winw = w - winx*winw;
		glViewport(x+winx,y+winy,winw,winh);
		anims[i]->draw(time,difftime);

		printName(anims[i]->getAlgorithm()->description, winh);

		if (n>1) drawBorder();
	}
	glViewport(x,y,w,h);
//	counter.count(time);
//	counter.draw();
	SDL_GL_SwapBuffers();

	SDL_Delay(1);
	if (paused) SDL_Delay(50);

}




void drawGraphs(Anim **anims, vector<threadFunc> &funcs) {
	bool running = false;
	bool allInserted = true;
	for (int i = 0; anims[i]; i++) {
		if (!anims[i]->isDone()) running = true;
		else if (anims[i]->getThread() != NULL) {
			cothread_join(anims[i]->getThread());
			anims[i]->setThread(NULL);
		}
		if (anims[i]->getType()==Anim::TREE) {
			TreeAnim *ta = (TreeAnim*) anims[i];
			if (!ta->inserted) allInserted = false;
		}
	}
	if (allInserted) {
		for (int i = 0; anims[i]; i++) {
			if (anims[i]->getType()==Anim::TREE) {
				TreeAnim *ta = (TreeAnim*) anims[i];
				ta->allInserted = true;
			}
		}
	}
	if (!running) {
		int firstgraph = -1;
		int firsttree = -1;
		for (int i = 0; anims[i]; i++) {
			if (anims[i]->getType()==Anim::GRAPH) {
				GraphAnim *ga = (GraphAnim*) anims[i];
				if (firstgraph<0) {
					ga->graph->initValues(order);
					firstgraph = i;
				} else {
					ga->graph->copyValues(((GraphAnim*)anims[firstgraph])->graph);
				}
			} else if (anims[i]->getType()==Anim::TREE) {
				TreeAnim *ta = (TreeAnim*) anims[i];
				if (firsttree<0) {
					ta->initValues(order,treeelements);
					firsttree = i;
				} else {
					ta->copyValues((TreeAnim*)anims[firsttree]);
				}
			}
		}
		for (int i = 0; anims[i]; i++) {
			anims[i]->setDone(false);
			anims[i]->setThread(cothread_create(funcs[i], anims[i]));
		}
	}
	for (int i = 0; anims[i]; i++) {
		if (anims[i]->updateAnims()) {
			if (!anims[i]->isDone()) {
				anims[i]->swapThreads();
			}
		}
	}
	drawFrame(anims);
}




void showGui(vector<threadFunc> &funcs, vector<Anim::AnimType> &types, DataOrder &order, vector<const char*> &modes, int &mode, bool &fullscreen, bool nofullscreen, int &sortelements, int &treeelements);

int main(int argc, char *argv[]) {
//	threadFunc funcs[10];
	vector<threadFunc> funcs;
	vector<Anim::AnimType> types;
	bool help = false;
	bool gui = false;
	int width = 640, height = 480;

	for (int i = 1; i < argc; i++) {
		if (!strcmp(argv[i],"-h")) {
			help = true;
			break;
		} else if (!strcmp(argv[i],"-g")) {
			gui = true;
			break;
		} else if (!strcmp(argv[i],"sortelements")) {
			if (i+1<argc) {
				i++;
				sortelements = atoi(argv[i]);
			}
		} else if (!strcmp(argv[i],"treeelements")) {
			if (i+1<argc) {
				i++;
				treeelements = atoi(argv[i]);
			}
		} else {
			struct algorithm *a = sortalgorithms;
			bool found = false;
			while (a->func!=NULL) {
				if (!strcmp(argv[i],a->name)) {
					funcs.push_back(a->func);
					types.push_back(Anim::GRAPH);
					found = true;
					break;
				}
				a++;
			}
			if (!found) {
				a = treealgorithms;
				while (a->func!=NULL) {
					if (!strcmp(argv[i],a->name)) {
						funcs.push_back(a->func);
						types.push_back(Anim::TREE);
						found = true;
						break;
					}
					a++;
				}
			}
			if (!found) {
				struct OrderDesc *od = orders;
				while (od->name!=NULL) {
					if (!strcmp(argv[i],od->name)) {
						order = od->order;
						found = true;
						break;
					}
					od++;
				}
			}
		}
	}

	if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE)<0) {
		fprintf(stderr,"%s\n",SDL_GetError());
		return 1;
	}
	if (atexit(SDL_Quit)) {
		fprintf(stderr,"atexit(...)\n");
		SDL_Quit();
		return 1;
	}

//	if (funcs.size()==0 || help) {
	if (!help && (funcs.size()==0 || gui)) {
		SDL_Rect **modes = NULL;
		modes = SDL_ListModes(NULL, SDL_OPENGL | SDL_FULLSCREEN);
		vector<const char *> modevect;
		bool nofullscreen = false;
		if (modes == (SDL_Rect**)0 || modes == (SDL_Rect**)-1) {
			if (modes == (SDL_Rect**)0)
				nofullscreen = true;
			modevect.push_back("640x480");
			modevect.push_back("800x600");
			modevect.push_back("1024x768");
			modevect.push_back("1280x1024");
			modevect.push_back("1600x1200");
		} else {
			for (int i = 0; modes[i]; i++) {
				if (i==0 || modes[i]->w!=modes[i-1]->w || modes[i]->h!=modes[i]->h) {
					char buffer[100];
					sprintf(buffer,"%dx%d",modes[i]->w,modes[i]->h);
					modevect.push_back(strdup(buffer));
				}
			}
		}

		int mode = 0;
		showGui(funcs,types,order,modevect,mode,fullscreen,nofullscreen,sortelements,treeelements);
		if (funcs.size()>0) {
			sscanf(modevect[mode],"%dx%d",&width,&height);
			for (int i = 0; i < int(modevect.size()); i++)
				free((void*)modevect[i]);
		}
	}
	if (funcs.size()==0 || help) {
		printf("%s algorithm1 [algorithm2 [algorithm3] ...]\n",argv[0]);
		printf("\n");
		printf("sort algorithms:\n");
		struct algorithm *a = sortalgorithms;
		while (a->func!=NULL) {
			printf("\t%s\t%s\n",a->name,a->description);
			a++;
		}
		printf("tree algorithms:\n");
		a = treealgorithms;
		while (a->func!=NULL) {
			if (strlen(a->name)<8)
				printf("\t%s\t\t%s\n",a->name,a->description);
			else
				printf("\t%s\t%s\n",a->name,a->description);
			a++;
		}
		printf("data formats:\n");
		struct OrderDesc *od = orders;
		while (od->name!=NULL) {
			printf("\t%s%s\n",od->name,od->order==order?" [default]":"");
			od++;
		}
		printf("other:\n");
		printf("\tsortelements n\tnumber of elements to sort\n");
		printf("\ttreeelements n\tnumber of elements to insert into trees\n");
		return 0;
	}

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
	SDL_WM_SetCaption("Data structures & Algorithms","Data structures & Algorithms");
	flags = SDL_OPENGL|SDL_RESIZABLE|(fullscreen?SDL_FULLSCREEN:0);
	screen = SDL_SetVideoMode(width,height,0,flags);
	if (screen==NULL) {
		fprintf(stderr,"%s\n",SDL_GetError());
		return 1;
	}

	vector<const struct algorithm*> algs;
	for (int i = 0; i < int(funcs.size()); i++) {
		struct algorithm* a = NULL;
		if (types[i]==Anim::GRAPH)
			a = sortalgorithms;
		else if (types[i]==Anim::TREE)
			a = treealgorithms;

		while (a->func!=NULL) {
			if (a->func == funcs[i]) {
				algs.push_back(a);
				break;
			}
			a++;
		}
	}

	srand(time(NULL));


	init();
	resize(screen->w,screen->h);
	LCDDisplay::precalc();


	if (sortelements<5) sortelements = 5;
	if (treeelements<5) treeelements = 5;

	active = true;
	starttime = SDL_GetTicks();
	lasttime = 0;
	turbo = false;
	slowmotion = false;
	paused = false;

	Anim **anims;

	anims = new Anim*[funcs.size()+1];

	for (int i = 0; i < int(funcs.size()); i++) {
		if (types[i]==Anim::GRAPH)
			anims[i] = new GraphAnim(new Graph(sortelements));
		else if (types[i]==Anim::TREE)
			anims[i] = new TreeAnim(new Forest());
		anims[i]->setAlgorithm(algs[i]);
	}
	anims[funcs.size()] = NULL;

	while (!done) {
		drawGraphs(anims,funcs);
	}

	for (int i = 0; anims[i]; i++) {
		if (anims[i]->getThread() != NULL) {
			anims[i]->setDone(true);
			cothread_join(anims[i]->getThread());
			anims[i]->setThread(NULL);
		}
		delete anims[i];
	}
	delete [] anims;

	deleteFont();

	return 0;
}


