/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <SDL_opengl.h>
#include "fps.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

FPSCounter::FPSCounter(int u) {
	update = u;
	lasttime = 0;
	lastframes = 0;
	totframes = 0;
	curframes = 0;
	curtime = 1;
	r = 0;
	g = 1;
	b = 0;
}

FPSCounter::~FPSCounter() {
}

void FPSCounter::count(int time) {
	if (lastframes==0) {
		curframes = totframes;
		curtime = time;
	}
	if (time-lasttime>update) {
		curframes = totframes - lastframes;
		curtime = time - lasttime;
		lasttime = time;
		lastframes = totframes;
	}

	totframes++;
}

void FPSCounter::draw(int width, int height) {
	char buffer[15];
	if (curtime==0) curtime = 1;
	sprintf(buffer,"%04d fps",1000*curframes/curtime);
	LCDDisplay::draw(1, 1, 6, 0.8, 0, buffer, r, g, b, width, height);
}

void FPSCounter::setColor(float r, float g, float b) {
	this->r = r;
	this->g = g;
	this->b = b;
}

namespace LCDDisplay {
//private:
	const int TOP = 0x01, MIDDLE = 0x02, BOTTOM = 0x04;
	const int TOPLEFT = 0x08, BOTTOMLEFT = 0x10, TOPRIGHT = 0x20, BOTTOMRIGHT = 0x40;
	GLuint list = 0;
	bool motionblur = false;
	void init() {
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0,100,100,0,-1,1);

		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_TEXTURE_2D);
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	}
	void finish() {
		glPopAttrib();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}
	int findElements(char c) {
		int out = 0;
		switch (c) {
		case '0': out = ~MIDDLE; break;
		case '1': out = TOPRIGHT|BOTTOMRIGHT; break;
		case '2': out = ~(TOPLEFT|BOTTOMRIGHT); break;
		case '3': out = ~(TOPLEFT|BOTTOMLEFT); break;
		case '4': out = ~(TOP|BOTTOM|BOTTOMLEFT); break;
		case '5': out = ~(TOPRIGHT|BOTTOMLEFT); break;
		case '6': out = ~TOPRIGHT; break;
		case '7': out = TOP|TOPRIGHT|BOTTOMRIGHT; break;
		case '8': out = ~0; break;
		case '9': out = ~BOTTOMLEFT; break;
		case 'f': out = ~(TOPRIGHT|BOTTOMRIGHT|BOTTOM); break;
		case 'p': out = ~(BOTTOMRIGHT|BOTTOM); break;
		case 's': out = ~(TOPRIGHT|BOTTOMLEFT); break;
		default: break;
		}
		return out;
	}
	void drawElement(int elements, int which, float r, float g, float b) {
		if (elements&which) {
			glColor4f(r,g,b,1);
		} else {
			if (motionblur) {
//				glColor4f(0.2*r,0.2*g,0.2*b,0.02);
				return;
			} else {
				glColor4f(r,g,b,0.07);
			}
		}

		glPushMatrix();
		switch (which) {
		case TOPLEFT: glTranslatef(1,2,0); break;
		case TOPRIGHT: glTranslatef(9,2,0); break;
		case BOTTOMLEFT: glTranslatef(1,10,0); break;
		case BOTTOMRIGHT: glTranslatef(9,10,0); break;
		case TOP: glTranslatef(2,1,0); glRotatef(-90,0,0,1); break;
		case BOTTOM: glTranslatef(2,17,0); glRotatef(-90,0,0,1); break;
		case MIDDLE: glTranslatef(2,9,0); glRotatef(-90,0,0,1); break;
		}

		glCallList(list);
/*		glBegin(GL_POLYGON);
		glVertex2i(0,0);
		glVertex2i(-1,1);
		glVertex2i(-1,5);
		glVertex2i(0,6);
		glVertex2i(1,5);
		glVertex2i(1,1);
		glEnd();
*/		glPopMatrix();
	}
	void drawChar(char c, float r, float g, float b) {
		int elements = findElements(c);
		drawElement(elements,TOP,r,g,b);
		drawElement(elements,TOPLEFT,r,g,b);
		drawElement(elements,TOPRIGHT,r,g,b);
		drawElement(elements,MIDDLE,r,g,b);
		drawElement(elements,BOTTOM,r,g,b);
		drawElement(elements,BOTTOMLEFT,r,g,b);
		drawElement(elements,BOTTOMRIGHT,r,g,b);
	}
//public:
	void precalc() {
		if (list!=0) glDeleteLists(list,1);
		list = glGenLists(1);
		glNewList(list, GL_COMPILE);
		glBegin(GL_POLYGON);
		glVertex2i(0,0);
		glVertex2i(-1,1);
		glVertex2i(-1,5);
		glVertex2i(0,6);
		glVertex2i(1,5);
		glVertex2i(1,1);
		glEnd();
		glEndList();
	}
	void setMotionBlur(bool b) {
		motionblur = b;
	}
	void draw(float x, float y, float h, float wh, float skew, const char *str, float r, float g, float b, int width, int height) {
		GLint viewport[4];
		glGetIntegerv(GL_VIEWPORT,viewport);
		if (width == 0) width = viewport[2];
		if (height == 0) height = viewport[3];
		float w = float(height)/width*wh*h;
		init();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(x,y,0);
		GLfloat matrix[16];
		glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
		matrix[4] = -skew;
		glLoadMatrixf(matrix);
		glScalef(w/11,h/19,1);
		while (*str!='\0') {
			drawChar(*str,r,g,b);
			str++;
			glTranslatef(12,0,0);
		}
		glPopMatrix();
		finish();
	}
}

