/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL.h>
#include <SDL_opengl.h>
#include <math.h>
#include "graph.h"
#include "vect.h"
#include "camera.h"
#include "box.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif


Graph::Graph(int n) {
	GLfloat hp[] = { 0, 0, 0 };
	GLfloat hs[] = { -1, -1, -1 };
	memcpy(highlightPos,hp,sizeof(GLfloat)*3);
	memcpy(highlightSize,hs,sizeof(GLfloat)*3);
	memcpy(zoomPos,hp,sizeof(GLfloat)*3);
	memcpy(zoomSize,hs,sizeof(GLfloat)*3);
	zoomHighlight = true;


	numValues = n;
	max = numValues;
	values = new int[numValues];

	colors = new GLfloat[numValues*3];
	GLfloat blue[] = { 0, 0, 1 };
	memcpy(defaultColor,blue,sizeof(GLfloat)*3);
	for (int i = 0; i < numValues; i++)
		memcpy(&colors[3*i],defaultColor,sizeof(GLfloat)*3);

	hidden.insert(hidden.begin(), numValues, false);
}

Graph::~Graph() {
	delete [] values;
	delete [] colors;
}


void Graph::setMax(int m) {
	max = m;
}

void Graph::initValues(DataOrder order) {
	for (int i = 0; i < numValues; i++)
		values[i] = i;
	if (order==RANDOM) {
		for (int i = 0; i < numValues; i++) {
			int pos = i+rand()%(numValues-i);
			int tmp = values[pos];
			values[pos] = values[i];
			values[i] = tmp;
		}
	} else if (order==REVERSE) {
		for (int i = 0; i < numValues; i++)
			values[i] = numValues-i-1;
	} else if (order==SEMISORTED) {
		int nswap = numValues/5 + rand()%(numValues/5);
		for (int i = 0; i < nswap; i++) {
			int a = rand()%numValues;
			int diff = rand()%(numValues/3)-(numValues/6);
			int b = a+diff;
			if (b<0) b = 0;
			if (b>=numValues) b = numValues-1;

			int tmp = values[a];
			values[a] = values[b];
			values[b] = tmp;
		}
		int nshift = numValues/10 + rand()%(numValues/10);
		for (int i = 0; i < nshift; i++) {
			int a = rand()%numValues;
			int b = rand()%numValues;
			if (b<a) {
				int c = b;
				b = a;
				a = c;
			}

			if (rand()%2) {
				int tmp = values[a];
				for (int i = a; i < b; i++)
					values[i] = values[i+1];
				values[b] = tmp;
			} else {
				int tmp = values[b];
				for (int i = b; i > a; i--)
					values[i] = values[i-1];
				values[a] = tmp;
			}
		}
	}
}

void Graph::copyValues(Graph *g) {
	for (int i = 0; i < numValues; i++)
		values[i] = g->values[i];
}

void Graph::resize(int n) {
	int *newval = new int[n];
	GLfloat *newcolors = new GLfloat[n*3];
	int copy = (n<numValues)?n:numValues;
	memcpy(newval,values,sizeof(int)*copy);
	memcpy(newcolors,colors,sizeof(GLfloat)*3*copy);
	for (int i = copy; i < n; i++) {
		newval[i] = 0;
		memcpy(&newcolors[3*i],defaultColor,sizeof(GLfloat)*3);
	}
	if (copy<numValues) {
		hidden.erase(hidden.begin()+n,hidden.end());
	} else {
		hidden.insert(hidden.end(), n-numValues, true);
	}
	numValues = n;
	delete [] values;
	delete [] colors;
	values = newval;
	colors = newcolors;
}




void Graph::setColor(int n, GLfloat *c) {
	memcpy(&colors[3*n],c,sizeof(GLfloat)*3);
}

void Graph::getColor(int n, GLfloat *c) {
	memcpy(c,&colors[3*n],sizeof(GLfloat)*3);
}

void Graph::getDefaultColor(GLfloat *c) {
	memcpy(c,defaultColor,sizeof(GLfloat)*3);
}

void Graph::resetColor(int n) {
	memcpy(&colors[3*n],defaultColor,sizeof(GLfloat)*3);
}

void Graph::setZoomHighlight(bool z) {
	zoomHighlight = z;
}

void Graph::hide(int n, bool h) {
	hidden[n] = h;
}


void Graph::setMovable(int id, GLfloat *pos, GLfloat *size, GLfloat *color) {
	int n = movables.size();
	Movable *m = NULL;
	for (int i = 0; i < n; i++) {
		if (movables[i].id == id) {
			m = &movables[i];
			break;
		}
	}
	if (m==NULL) {
		movables.push_back(Movable());
		m = &movables[n];
	}
	m->id = id;
	memcpy(m->pos,pos,sizeof(GLfloat)*3);
	if (size!=NULL)
		memcpy(m->size,size,sizeof(GLfloat)*3);
	else
		getSize(m->id,m->size);
	if (size!=NULL)
		memcpy(m->color,color,sizeof(GLfloat)*3);
	else
		memcpy(m->color,&colors[3*m->id],sizeof(GLfloat)*3);
}

void Graph::getMovable(int id, GLfloat *pos, GLfloat *size, GLfloat *color) {
	int n = movables.size();
	Movable *m = NULL;
	for (int i = 0; i < n; i++) {
		if (movables[i].id == id) {
			m = &movables[i];
			break;
		}
	}
	if (m==NULL) return;
	memcpy(pos,m->pos,sizeof(GLfloat)*3);
	memcpy(size,m->size,sizeof(GLfloat)*3);
	memcpy(color,m->color,sizeof(GLfloat)*3);
}

void Graph::removeMovable(int id) {
	int n = movables.size();
	for (int i = 0; i < n; i++) {
		if (movables[i].id==id) {
			movables.erase(movables.begin()+i);
			return;
		}
	}
}


void Graph::getPos(int i, GLfloat *pos) {
	pos[0] = i*0.2;
	pos[1] = 0;
	pos[2] = 0;
}

void Graph::getSize(int i, GLfloat *size) {
	size[0] = 0.15;
	size[1] = 10.*values[i]/max+0.3;
	size[2] = 0.2;
}

void Graph::setHighlight(GLfloat *pos, GLfloat *size) {
	memcpy(highlightPos,pos,sizeof(GLfloat)*3);
	memcpy(highlightSize,size,sizeof(GLfloat)*3);
}

void Graph::getHighlight(GLfloat *pos, GLfloat *size) {
	memcpy(pos,highlightPos,sizeof(GLfloat)*3);
	memcpy(size,highlightSize,sizeof(GLfloat)*3);
}

void Graph::removeHighlight() {
	highlightSize[0] = -1;
}

void Graph::setZoom(GLfloat *pos, GLfloat *size) {
	memcpy(zoomPos,pos,sizeof(GLfloat)*3);
	memcpy(zoomSize,size,sizeof(GLfloat)*3);
}

void Graph::getZoom(GLfloat *pos, GLfloat *size) {
	memcpy(pos,zoomPos,sizeof(GLfloat)*3);
	memcpy(size,zoomSize,sizeof(GLfloat)*3);
}

void Graph::removeZoom() {
	zoomSize[0] = -1;
}

void Graph::draw(int time, int difftime) {
	GLfloat lightSpecular[] = { 0, 0, 0, 0 };
	GLfloat lightDiffuse[] = { 0.4, 0.4, 0.4, 1 };
	GLfloat lightAmbient[] = { 0.1, 0.1, 0.1, 1 };
	GLfloat light0Pos[] = { 0, 5, 10, 1 };
	GLfloat light1Pos[] = { 20, 5, 10, 1 };
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glEnable(GL_COLOR_MATERIAL);
	glLightfv(GL_LIGHT0, GL_SPECULAR, lightSpecular);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
	glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
	glLightfv(GL_LIGHT1, GL_SPECULAR, lightSpecular);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightDiffuse);
	glLightfv(GL_LIGHT1, GL_AMBIENT, lightAmbient);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	GLfloat viewport[4];
	glGetFloatv(GL_VIEWPORT, viewport);
	gluPerspective(45,viewport[2]/viewport[3],0.1,100);

	GLfloat upleft[3], upright[3], downleft[3];
	GLfloat *pos = zoomPos, *size = zoomSize;
	if (zoomHighlight) {
		pos = highlightPos;
		size = highlightSize;
	}
	if (size[0]>0) {
		Vect_copy3f(downleft,pos);
		downleft[2] += 2;
		Vect_copy3f(upleft,downleft);
		upleft[1] += size[1];
		Vect_copy3f(upright,upleft);
		upright[0] += size[0];
	} else {
		GLfloat ul[] = { -0.3, 10.5, 1 }, ur[] = { numValues*0.2 + 0.3, 10.5, 1 }, dl[] = { -0.3, 0, 1 };
		Vect_copy3f(upleft,ul);
		Vect_copy3f(upright,ur);
		Vect_copy3f(downleft,dl);
	}
	light1Pos[0] = upright[0];
	light0Pos[0] = upleft[0];
	glLightfv(GL_LIGHT0, GL_POSITION, light0Pos);
	glLightfv(GL_LIGHT1, GL_POSITION, light1Pos);
	setCamera(upleft,upright,downleft,45,true);


	glMatrixMode(GL_MODELVIEW);
/*	if (reset && time-morphStart<300) {
		float pos = float(time-morphStart)/300;
		GLfloat matrix[16];
		resetMorph.rotate(pos,matrix);
		glLoadMatrixf(matrix);
	} else if (reset) {
		reset = false;
		glLoadMatrixf(origMatrix);
	}
*/

	glBegin(GL_QUADS);
	for (int i = 0; i < numValues; i++) {
		if (hidden[i]) continue;
		GLfloat c[3];
		GLfloat pos[3];
		GLfloat size[3];

		getPos(i,pos);
		getSize(i,size);
		memcpy(c,&colors[3*i],sizeof(GLfloat)*3);

		Box b(c[0],c[1],c[2],1, pos[0],pos[1],pos[2], size[0],size[1],size[2]);
		drawBox(b,false);
	}
	int n = movables.size();
	for (int i = 0; i < n; i++) {
		Movable &m = movables[i];
		Box b(m.color[0],m.color[1],m.color[2],1, m.pos[0],m.pos[1],m.pos[2], m.size[0],m.size[1],m.size[2]);
		drawBox(b,false);
	}
	glEnd();
	vector<Box> boxes; // for alpha boxes
	if (highlightSize[0]>0) {
		GLfloat *pos = highlightPos;
		GLfloat *size = highlightSize;
		boxes.push_back(Box(1,1,0,0.4, pos[0], pos[1], pos[2], size[0], size[1], size[2]));
	}

	drawBoxes(boxes);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

}


