/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include <vector>
#include <math.h>
#include "graphanim.h"
#include "graph.h"
#include "vect.h"
#include "spline.h"

using std::vector;

enum Type { FADE, MOVE, HIGHLIGHT, ZOOM };
class GraphAnimItem {
public:
	int starttime;
	int length;
	int type;
	int n;
	Spline3 move;
	Spline3 size;
	GLfloat color[3];
	GLfloat moveSize[3];
};


GraphAnim::GraphAnim(Graph *g) {
	graph = g;
}

GraphAnim::~GraphAnim() {
	delete graph;
}


void GraphAnim::markRed(int n) {
	GLfloat curColor[3], newColor[] = { 1, 0, 0 };
	graph->getColor(n,curColor);

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = FADE;
	a.n = n;
	a.move = Spline3(NULL,curColor,newColor,NULL);
	anims.push_back(a);
}

void GraphAnim::markGreen(int n) {
	GLfloat curColor[3], newColor[] = { 0, 1, 0 };
	graph->getColor(n,curColor);

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = FADE;
	a.n = n;
	a.move = Spline3(NULL,curColor,newColor,NULL);
	anims.push_back(a);
}

void GraphAnim::clearMark(int n) {
	GLfloat curColor[3], newColor[3];
	graph->getColor(n,curColor);
	graph->getDefaultColor(newColor);

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = FADE;
	a.n = n;
	a.move = Spline3(NULL,curColor,newColor,NULL);
	anims.push_back(a);
}

void GraphAnim::clearMarks() {
	GLfloat defaultColor[3];
	graph->getDefaultColor(defaultColor);
	for (int i = 0; i < graph->numValues; i++) {
		GLfloat color[3];
		graph->getColor(i,color);
		if (fabs(color[0]-defaultColor[0])>0.01 || fabs(color[1]-defaultColor[1])>0.01 || fabs(color[2]-defaultColor[2])>0.01) {
			clearMark(i);
		}
	}
}



bool GraphAnim::updateAnim(GraphAnimItem &a) {
	if (a.starttime>animtime) return false;
	int time = animtime - a.starttime;
	float t = float(time)/a.length;
	bool done = false;
	if (t>1) {
		done = true;
		t = 1;
	}
	if (a.type==FADE) {
		GLfloat color[3];
		a.move.calc(t,color);
		graph->setColor(a.n,color);
	} else if (a.type==MOVE) {
		GLfloat pos[3];
		a.move.calc(t,pos);
		graph->setMovable(a.n,pos,a.moveSize,a.color);
	} else if (a.type==HIGHLIGHT || a.type==ZOOM) {
		GLfloat pos[3], size[3];
		a.move.calc(t,pos);
		a.size.calc(t,size);
		if (a.type==HIGHLIGHT) graph->setHighlight(pos,size);
		else graph->setZoom(pos,size);
	}
	return done;
}


bool GraphAnim::updateAnims() {
	int n = anims.size();
	for (int i = 0; i < n; i++) {
		if (updateAnim(anims[i])) {
			anims.erase(anims.begin()+i);
			i--;
			n--;
		}
	}
	return n==0;
}

void GraphAnim::drawAnim(int msec) {
/*	while (anims.size()>0) {
		int n = anims.size();
		for (int i = 0; i < n; i++) {
			if (updateAnim(anims[i])) {
				anims.erase(anims.begin()+i);
				i--;
				n--;
			}
		}
		drawFrame();
	}
	int time = animtime;
	while (animtime-time<msec) {
		drawFrame();
	}
*/
	int starttime = animtime;
	do {
		swapThreads();
	} while (animtime<starttime+msec);
}

void GraphAnim::clearAnims() {
	anims.clear();
}

void GraphAnim::draw(int time, int difftime) {
	graph->draw(time,difftime);
}


void GraphAnim::animSwap(int a, int b, bool doswap) {
	if (a==b) return;
	GLfloat origa[3], origb[3], waypointa[3], waypointb[3];
	graph->getPos(a,origa);
	graph->getPos(b,origb);
	Vect_copy3f(waypointa,origa);
	waypointa[2] += 1.5;
	Vect_copy3f(waypointb,origa);
	Vect_add3f(waypointb,origb);
	Vect_mult3f(waypointb,0.5);
	waypointb[2] += 1.5;

	GraphAnimItem anim;
	anim.starttime = animtime;
	anim.length = 300;
	anim.n = a;
	anim.type = MOVE;
	anim.move = Spline3(NULL,origa,waypointa,NULL);
	graph->getColor(a,anim.color);
	graph->getSize(a,anim.moveSize);
	anims.push_back(anim);

	graph->hide(a, true);
	drawAnim();

	anim.starttime = animtime;
	anim.length = 150;
	anim.n = b;
	anim.move = Spline3(NULL,origb,waypointb,origa);
	graph->getColor(b,anim.color);
	graph->getSize(b,anim.moveSize);
	anims.push_back(anim);

	anim.starttime += 150;
	anim.move = Spline3(origb,waypointb,origa,NULL);
	anims.push_back(anim);

	anim.starttime += 150;
	anim.length = 300;
	anim.n = a;
	anim.move = Spline3(NULL,waypointa,origb,NULL);
	graph->getColor(a,anim.color);
	graph->getSize(a,anim.moveSize);
	anims.push_back(anim);

	graph->hide(b, true);
	drawAnim();

	if (doswap) {
		int tmp = graph->values[a];
		graph->values[a] = graph->values[b];
		graph->values[b] = tmp;
	}

	graph->hide(a, false);
	graph->hide(b, false);
	graph->removeMovable(a);
	graph->removeMovable(b);
}

void GraphAnim::moveForward(int n) {
	graph->hide(n,true);

	GLfloat oldPos[3], newPos[3];
	graph->getPos(n,oldPos);
	memcpy(newPos,oldPos,sizeof(GLfloat)*3);
	newPos[2] += 1.5;

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.n = n;
	a.type = MOVE;
	a.move = Spline3(NULL,oldPos,newPos,NULL);
	graph->getColor(n,a.color);
	graph->getSize(n,a.moveSize);
	anims.push_back(a);
	drawAnim();
}

void GraphAnim::slideRight(int n, bool docopy) {
	graph->hide(n,true);

	GLfloat oldPos[3], newPos[3];
	graph->getPos(n,oldPos);
	graph->getPos(n+1,newPos);

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 150;
	a.n = n;
	a.type = MOVE;
	a.move = Spline3(NULL,oldPos,newPos,NULL);
	graph->getColor(n,a.color);
	graph->getSize(n,a.moveSize);
	anims.push_back(a);
	drawAnim();
	if (docopy)
		graph->values[n+1] = graph->values[n];
	graph->hide(n+1,false);
	graph->setColor(n+1,a.color);
	graph->removeMovable(n);
}

void GraphAnim::moveTo(int n, int n2, bool docopy) {
	graph->hide(n,true);

	GLfloat oldPos[3], newPos[3];
	graph->getPos(n,oldPos);
	graph->getPos(n2,newPos);

	GLfloat waypoint[3];
	Vect_copy3f(waypoint,oldPos);
	Vect_add3f(waypoint,newPos);
	Vect_mult3f(waypoint,0.5);
	waypoint[2] += 1.5;

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 150;
	a.n = n;
	a.type = MOVE;
	a.move = Spline3(NULL,oldPos,waypoint,newPos);
	graph->getColor(n,a.color);
	graph->getSize(n,a.moveSize);
	anims.push_back(a);

	a.starttime += 150;
	a.move = Spline3(oldPos,waypoint,newPos,NULL);
	anims.push_back(a);
	drawAnim();

	if (docopy)
		graph->values[n2] = graph->values[n];
	graph->hide(n2,false);
	graph->setColor(n2,a.color);
	graph->removeMovable(n);
}

void GraphAnim::moveBackTo(int n, int n2) {
	GLfloat oldPos[3], newPos[3];

	GraphAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.n = n;
	a.type = MOVE;
	graph->getMovable(n,oldPos,a.moveSize,a.color);
	graph->getPos(n2,newPos);
	a.move = Spline3(NULL,oldPos,newPos,NULL);
	anims.push_back(a);
	drawAnim();

	graph->hide(n2,false);
	graph->setColor(n2,a.color);
	graph->removeMovable(n);
}

void GraphAnim::zoomHighlight(int type, int start, int end) {
	GLfloat newPos[3], newSize[3];
	GLfloat oldPos[3], oldSize[3];
	if (type==HIGHLIGHT) graph->getHighlight(oldPos,oldSize);
	else graph->getZoom(oldPos,oldSize);
	graph->getPos(start,newPos);
	newPos[0] -= 0.01;
	newPos[1] -= 0.3;
//	newPos[2] -= 0.05;
	graph->getPos(end+1,newSize);
	Vect_sub3f(newSize,newPos);
	newPos[2] -= 0.05;
//	newSize[0] += 0.01;
	newSize[1] += 10.3;

	if (oldSize[0]<0) {
		// fade in?
		if (type==HIGHLIGHT) graph->setHighlight(newPos,newSize);
		else graph->setZoom(newPos,newSize);
//		memcpy(graph->highlightPos,newPos,sizeof(GLfloat)*3);
//		memcpy(graph->highlightSize,newSize,sizeof(GLfloat)*3);
		return;
	}

	GraphAnimItem a;
	a.type = type;
	a.move = Spline3(NULL,oldPos,newPos,NULL);
	a.size = Spline3(NULL,oldSize,newSize,NULL);
	a.starttime = animtime;
	a.length = 500;
	anims.push_back(a);
	drawAnim();
}

void GraphAnim::highlight(int start, int end) {
	zoomHighlight(HIGHLIGHT,start,end);
}

void GraphAnim::zoom(int start, int end) {
	zoomHighlight(ZOOM,start,end);
}

void GraphAnim::clearZoom() {
	graph->removeZoom();
}

void GraphAnim::clearHighlight() {
	graph->removeHighlight();
}

void GraphAnim::swapNeighbors(int n) {
	int val = graph->values[n+1];
	moveForward(n+1);
	slideRight(n,true);
	graph->values[n] = val;
	moveBackTo(n+1,n);
}

