/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <vector>
#include "algorithms.h"

using std::vector;

#ifdef HAVE_GTK
#include <gtk/gtk.h>



static GtkWidget *ok;
static GtkWidget *combo;
static GtkWidget *fullscreencheck;
static GtkWidget *sortwidget;
static GtkWidget *treewidget;

static DataOrder curorder = RANDOM;
static vector<threadFunc> localfuncs;
static vector<Anim::AnimType> localtypes;
static bool done = false;
static bool localfullscreen = false;
static int localmode = 0;
static int localsortelements;
static int localtreeelements;

struct AlgorithmButton {
	struct algorithm *a;
	Anim::AnimType type;
	GtkWidget *button;
};

static vector<AlgorithmButton> buttons;



static void destroy(GtkWidget *widget, gpointer data) {
	gtk_main_quit();
	if (!done) exit(0);
}

static void click(GtkWidget *widget, gpointer data) {
	if (widget==ok) {
		int n = buttons.size();
		for (int i = 0; i < n; i++) {
			if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttons[i].button))) {
				localfuncs.push_back(buttons[i].a->func);
				localtypes.push_back(buttons[i].type);
			}
		}
		localmode = gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
		localfullscreen = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(fullscreencheck))==TRUE;
		localsortelements = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(sortwidget));
		localtreeelements = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(treewidget));
	}
	done = true;
	gtk_widget_destroy(GTK_WIDGET(data));
	if (widget!=ok) {
		exit(0);
	}
}


static void toggled(GtkToggleButton *button, gpointer data) {
	int n = buttons.size();
	bool checked = false;
	for (int i = 0; i < n; i++) {
		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttons[i].button)))
			checked = true;
	}
	gtk_widget_set_sensitive(ok, checked?TRUE:FALSE);
}

static void radiotoggled(GtkToggleButton *button, gpointer data) {
	int i = int(data);
	curorder = DataOrder(i);
}

static void showhelp(GtkWidget *widget, gpointer data) {
	GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(data), GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
		"Available keys:\n"
		"F1 - switch between fullscreen and window\n"
		"left arrow - slow motion\n"
		"right arrow - fast forward\n"
		"space bar - pause\n"
		"escape - exit");
	gtk_window_set_title(GTK_WINDOW(dialog), "Instructions");
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

void showGui(vector<threadFunc> &funcs, vector<Anim::AnimType> &types, DataOrder &order, vector<const char*> &modes, int &mode, bool &fullscreen, bool nofullscreen, int &sortelements, int &treeelements) {
	gtk_init(NULL,NULL);

	GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Data structures & Algorithms");
	g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(destroy), NULL);

	gtk_container_set_border_width(GTK_CONTAINER(window), 10);

	GtkContainer *hbox = GTK_CONTAINER(gtk_hbox_new(FALSE, 10));
	gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(hbox));

	GtkWidget *frame1 = gtk_frame_new("Sort algorithms");
	gtk_frame_set_shadow_type(GTK_FRAME(frame1), GTK_SHADOW_ETCHED_IN);
	gtk_container_add(hbox, frame1);

	GtkWidget *align1 = gtk_alignment_new(0.50, 0.50, 1, 1);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align1), 0, 5, 12, 5);
	gtk_container_add(GTK_CONTAINER(frame1), align1);
	GtkWidget *vbox1 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(align1), vbox1);

	GtkWidget *hbox1 = gtk_hbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);
	sortwidget = gtk_spin_button_new_with_range(5, 10000, 1);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(sortwidget), sortelements);
	gtk_box_pack_start(GTK_BOX(hbox1), sortwidget, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox1), gtk_label_new("elements"), FALSE, FALSE, 0);

	struct algorithm *a = sortalgorithms;
	while (a->func!=NULL) {
		GtkWidget *button = gtk_check_button_new_with_label(a->description);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), FALSE);
		g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggled), NULL);
		AlgorithmButton ab;
		ab.a = a;
		ab.type = Anim::GRAPH;
		ab.button = button;
		gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, 0);
		buttons.push_back(ab);
		a++;
	}

	GtkWidget *frame2 = gtk_frame_new("Tree algorithms");
	gtk_frame_set_shadow_type(GTK_FRAME(frame2), GTK_SHADOW_ETCHED_IN);
	gtk_container_add(hbox, frame2);

	GtkWidget *align2 = gtk_alignment_new(0.50, 0.50, 1, 1);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align2), 0, 5, 12, 5);
	gtk_container_add(GTK_CONTAINER(frame2), align2);
	GtkWidget *vbox2 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(align2), vbox2);

	GtkWidget *hbox2 = gtk_hbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
	treewidget = gtk_spin_button_new_with_range(5, 10000, 1);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(treewidget), treeelements);
	gtk_box_pack_start(GTK_BOX(hbox2), treewidget, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox2), gtk_label_new("elements"), FALSE, FALSE, 0);

	a = treealgorithms;
	while (a->func!=NULL) {
		GtkWidget *button = gtk_check_button_new_with_label(a->description);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), FALSE);
		g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggled), NULL);
		AlgorithmButton ab;
		ab.a = a;
		ab.type = Anim::TREE;
		ab.button = button;
		gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
		buttons.push_back(ab);
		a++;
	}

	GtkWidget *framebox = gtk_vbox_new(FALSE, 10);
	gtk_container_add(hbox, framebox);

	GtkWidget *frame3 = gtk_frame_new("Data order");
	gtk_frame_set_shadow_type(GTK_FRAME(frame3), GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(framebox), frame3);

	GtkWidget *align3 = gtk_alignment_new(0.50, 0.50, 1, 1);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align3), 0, 5, 12, 5);
	gtk_container_add(GTK_CONTAINER(frame3), align3);
	GtkWidget *vbox3 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(align3), vbox3);

	struct OrderDesc *od = orders;
	bool first = true;
	GSList *group = NULL;
	while (od->name!=NULL) {
		GtkWidget *button = gtk_radio_button_new_with_label(group,od->description);
		group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button));
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), first?TRUE:FALSE);
		if (first) curorder = od->order;
		first = false;
		g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(radiotoggled), gpointer(od->order));
		gtk_box_pack_start(GTK_BOX(vbox3), button, FALSE, FALSE, 0);
		od++;
	}

	GtkWidget *frame4 = gtk_frame_new("Display");
	gtk_frame_set_shadow_type(GTK_FRAME(frame4), GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(framebox), frame4);

	GtkWidget *align4 = gtk_alignment_new(0.50, 0.50, 1, 1);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align4), 0, 5, 12, 5);
	gtk_container_add(GTK_CONTAINER(frame4), align4);
	GtkWidget *vbox4 = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(align4), vbox4);

	combo = gtk_combo_box_new_text();
	int activemode = 0;
	for (int i = 0; i < int(modes.size()); i++) {
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo), modes[i]);
		if (!strcmp(modes[i],"640x480")) activemode = i;
	}
	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), activemode);
	gtk_box_pack_start(GTK_BOX(vbox4), combo, FALSE, FALSE, 0);
	fullscreencheck = gtk_check_button_new_with_label("Fullscreen");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fullscreencheck), FALSE);
	if (nofullscreen) gtk_widget_set_sensitive(fullscreencheck, FALSE);
	gtk_box_pack_start(GTK_BOX(vbox4), fullscreencheck, FALSE, FALSE, 0);



	GtkWidget *frame5 = gtk_frame_new("");
	gtk_frame_set_shadow_type(GTK_FRAME(frame5), GTK_SHADOW_NONE);
	gtk_container_add(hbox, frame5);

	GtkWidget *buttonbox = gtk_vbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttonbox), GTK_BUTTONBOX_START);
	gtk_box_set_spacing(GTK_BOX(buttonbox), 10);
	gtk_container_add(GTK_CONTAINER(frame5),buttonbox);

	ok = gtk_button_new_from_stock(GTK_STOCK_OK);
	gtk_widget_set_sensitive(ok, FALSE);
	g_signal_connect(G_OBJECT(ok), "clicked", G_CALLBACK(click), window);
	GtkWidget *cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	g_signal_connect(G_OBJECT(cancel), "clicked", G_CALLBACK(click), window);
	GtkWidget *help = gtk_button_new_from_stock(GTK_STOCK_HELP);
	g_signal_connect(G_OBJECT(help), "clicked", G_CALLBACK(showhelp), window);
	gtk_container_add(GTK_CONTAINER(buttonbox), ok);
	gtk_container_add(GTK_CONTAINER(buttonbox), cancel);
	gtk_container_add(GTK_CONTAINER(buttonbox), help);

	gtk_widget_show_all(window);
	gtk_main();

	order = curorder;
	funcs = localfuncs;
	types = localtypes;
	mode = localmode;
	fullscreen = localfullscreen;
	sortelements = localsortelements;
	treeelements = localtreeelements;

	buttons.clear();
}

#else

void showGui(vector<threadFunc> &funcs, vector<Anim::AnimType> &types, DataOrder &order, vector<const char*> &modes, int &mode, bool &fullscreen, bool nofullscreen, int &sortelements, int &treeelements) {
}

#endif

