/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <string.h>
#include <SDL_opengl.h>
#include <math.h>
#include "matrix.h"

class Matrix {
private:
	GLfloat *data; // in column major order
	int n;
public:
	Matrix(int n, const GLfloat *in);
	Matrix(int n, int row, int column, const GLfloat *in);
	Matrix(int n);
	Matrix(const Matrix &m);
	Matrix &operator=(const Matrix &m);
	~Matrix();
	float det();
	void print();
	Matrix operator*(const Matrix &m) const;
	Matrix invert();
	Matrix gaussInvert();
	Matrix gaussJordan();
	Matrix fakeInvert();
	void get(GLfloat *out);
	int size();
};

Matrix::Matrix(int n) {
	this->n = n;
	data = new GLfloat[n*n];
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++)
			data[i*n+j] = (i==j)?1:0;
}

Matrix::Matrix(int n, const GLfloat *in) {
	this->n = n;
	data = new GLfloat[n*n];
	memcpy(data,in,sizeof(GLfloat)*n*n);
}

Matrix::Matrix(int n, int row, int column, const GLfloat *in) {
	n--;
	this->n = n;
	data = new GLfloat[n*n];
	for (int i = 0; i < n+1; i++) {
		int k = i;
		if (k==column) continue;
		else if (k>column) k--;
		for (int j = 0; j < n+1; j++) {
			int l = j;
			if (l==row) continue;
			else if (l>row) l--;
			data[n*k+l] = in[(n+1)*i+j];
		}
	}
}

Matrix::Matrix(const Matrix &m) {
	n = m.n;
	data = new GLfloat[n*n];
	memcpy(data,m.data,sizeof(GLfloat)*n*n);
}

Matrix &Matrix::operator=(const Matrix &m) {
	delete [] data;
	n = m.n;
	data = new GLfloat[n*n];
	memcpy(data,m.data,sizeof(GLfloat)*n*n);
	return *this;
}

Matrix::~Matrix() {
	delete [] data;
}

float Matrix::det() {
	if (n==2) {
		return data[0]*data[3]-data[1]*data[2];
	} else {
		float d = 0;
		for (int i = 0; i < n; i++) {
			d += data[i]*Matrix(n,i,0,data).det()*(i%2==0?1:-1);
		}
		return d;
	}
}

void Matrix::print() {
	for (int i = 0; i < n; i++) {
		printf("[");
		for (int j = 0; j < n; j++) {
			printf("%2.2f",data[j*n+i]);
			if (j<n-1) printf(" ");
		}
		printf("]\n");
	}
	printf("\n");
}

Matrix Matrix::invert() {
	GLfloat out[16];
	float det = Matrix(n,data).det();
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			out[j*n+i] = Matrix(n,j,i,data).det()*((i+j)%2==0?1:-1)/det;
		}
	}
	return Matrix(n,out);
}

void Matrix::get(GLfloat *out) {
	memcpy(out,data,sizeof(GLfloat)*n*n);
}

int Matrix::size() {
	return n;
}

Matrix Matrix::gaussInvert() {
	Matrix A(n,data);
	int *vars = new int[n];
	for (int i = 0; i < n; i++)
		vars[i] = i;

	Matrix b(n);
	GLfloat *temp = new GLfloat[n];

	for (int col = 0; col < n-1; col++) {
		int maxrow = col;
		int maxcol = col;
		GLfloat maxval = fabs(A.data[maxrow+maxcol*n]);
		for (int r = col; r < n; r++) {
			for (int c = col; c < n; c++) {
				if (fabs(A.data[r*n+c])>maxval) {
					maxrow = r;
					maxcol = c;
					maxval = fabs(A.data[r+c*n]);
				}
			}
		}
		if (maxrow!=col) {
			// swap rows maxrow and col in A and b
			for (int i = 0; i < n; i++)
				temp[i] = A.data[col+i*n];
			for (int i = 0; i < n; i++)
				A.data[col+i*n] = A.data[maxrow+i*n];
			for (int i = 0; i < n; i++)
				A.data[maxrow+i*n] = temp[i];
			for (int i = 0; i < n; i++)
				temp[i] = b.data[col+i*n];
			for (int i = 0; i < n; i++)
				b.data[col+i*n] = b.data[maxrow+i*n];
			for (int i = 0; i < n; i++)
				b.data[maxrow+i*n] = temp[i];
		}
		if (maxcol!=col) {
			// swap cols maxcol and col in A and vars
			memcpy(temp,&A.data[col*n],sizeof(GLfloat)*n);
			memcpy(&A.data[col*n],&A.data[maxcol*n],sizeof(GLfloat)*n);
			memcpy(&A.data[maxcol*n],temp,sizeof(GLfloat)*n);
			int tmpvar = vars[col];
			vars[col] = vars[maxcol];
			vars[maxcol] = tmpvar;
		}
		for (int row = col+1; row < n; row++) {
			GLfloat fact = -A.data[row+col*n]/A.data[col+col*n];
			for (int i = 0; i < n; i++)
				A.data[row+i*n] = A.data[row+i*n] + fact*A.data[col+i*n];
			for (int i = 0; i < n; i++)
				b.data[row+i*n] = b.data[row+i*n] + fact*b.data[col+i*n];
		}
	}

	Matrix out(n);
	for (int vect = 0; vect < n; vect++) {
		for (int i = n-1; i >= 0; i--) {
			GLfloat val = b.data[i+vect*n];
			for (int col = n-1; col > i; col--)
				val -= A.data[i+col*n]*out.data[vars[col]+vect*n];
			val /= A.data[i+i*n];
			out.data[vars[i]+vect*n] = val;
		}
	}

	delete [] temp;
	delete [] vars;
	return out;
}

Matrix Matrix::gaussJordan() {
	Matrix A(n,data);
	int *vars = new int[n];
	for (int i = 0; i < n; i++)
		vars[i] = i;

	Matrix B(n);
	GLfloat *temp = new GLfloat[n];

	for (int col = 0; col < n-1; col++) {
		int maxrow = col;
		int maxcol = col;
		GLfloat maxval = fabs(A.data[maxrow+maxcol*n]);
		for (int r = col; r < n; r++) {
			for (int c = col; c < n; c++) {
				if (fabs(A.data[r*n+c])>maxval) {
					maxrow = r;
					maxcol = c;
					maxval = fabs(A.data[r+c*n]);
				}
			}
		}
		if (maxrow!=col) {
			// swap rows maxrow and col in A and b
			for (int i = 0; i < n; i++)
				temp[i] = A.data[col+i*n];
			for (int i = 0; i < n; i++)
				A.data[col+i*n] = A.data[maxrow+i*n];
			for (int i = 0; i < n; i++)
				A.data[maxrow+i*n] = temp[i];
			for (int i = 0; i < n; i++)
				temp[i] = B.data[col+i*n];
			for (int i = 0; i < n; i++)
				B.data[col+i*n] = B.data[maxrow+i*n];
			for (int i = 0; i < n; i++)
				B.data[maxrow+i*n] = temp[i];
		}
		if (maxcol!=col) {
			// swap cols maxcol and col in A and vars
			memcpy(temp,&A.data[col*n],sizeof(GLfloat)*n);
			memcpy(&A.data[col*n],&A.data[maxcol*n],sizeof(GLfloat)*n);
			memcpy(&A.data[maxcol*n],temp,sizeof(GLfloat)*n);
			int tmpvar = vars[col];
			vars[col] = vars[maxcol];
			vars[maxcol] = tmpvar;
		}
		for (int row = col+1; row < n; row++) {
			GLfloat fact = -A.data[row+col*n]/A.data[col+col*n];
			for (int i = 0; i < n; i++)
				A.data[row+i*n] = A.data[row+i*n] + fact*A.data[col+i*n];
			for (int i = 0; i < n; i++)
				B.data[row+i*n] = B.data[row+i*n] + fact*B.data[col+i*n];
		}
	}

	for (int col = n-1; col > 0; col--) {
		for (int row = col-1; row >= 0; row--) {
			GLfloat fact = -A.data[row+col*n]/A.data[col+col*n];
			for (int i = 0; i < n; i++)
				A.data[row+i*n] = A.data[row+i*n] + fact*A.data[col+i*n];
			for (int i = 0; i < n; i++)
				B.data[row+i*n] = B.data[row+i*n] + fact*B.data[col+i*n];
		}
	}

	Matrix out(n);
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++)
			out.data[vars[i]+j*n] = B.data[i+j*n]/A.data[i+i*n];
	}

	delete [] temp;
	delete [] vars;
	return out;
}

Matrix Matrix::operator*(const Matrix &m) const {
	if (n!=m.n) return Matrix(1);
	Matrix out(n);
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			GLfloat val = 0;
			for (int k = 0; k < n; k++)
				val += data[i+k*n]*m.data[k+j*n];
			out.data[i+j*n] = val;
		}
	}
	return out;
}

Matrix Matrix::fakeInvert() {
	Matrix rot(4);
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			rot.data[i+j*4] = data[j+i*4];
	Matrix trans(4);
	for (int i = 0; i < 3; i++)
		trans.data[i+3*4] = -data[i+3*4];
	return rot*trans;
}

void Matrix_invertf(GLfloat *out, const GLfloat *in) {
	Matrix(4,in).invert().get(out);
}

void Matrix_gaussInvertf(GLfloat *out, const GLfloat *in) {
	Matrix(4,in).gaussInvert().get(out);
}

void Matrix_gaussJordanf(GLfloat *out, const GLfloat *in) {
	Matrix(4,in).gaussJordan().get(out);
}

void Matrix_fakeInvertf(GLfloat *out, const GLfloat *in) {
	Matrix(4,in).fakeInvert().get(out);
}


void Matrix_multf(GLfloat *out, const GLfloat *m1, const GLfloat *m2) {
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			GLfloat f = 0;
			for (int k = 0; k < 4; k++)
				f += m1[k*4+j]*m2[i*4+k];
			out[i*4+j] = f;
		}
	}
}

void Matrix_multVect4f(GLfloat *out, const GLfloat *matrix, const GLfloat *vect) {
	out[0] = matrix[0*4+0]*vect[0]+matrix[1*4+0]*vect[1]+matrix[2*4+0]*vect[2]+matrix[3*4+0]*vect[3];
	out[1] = matrix[0*4+1]*vect[0]+matrix[1*4+1]*vect[1]+matrix[2*4+1]*vect[2]+matrix[3*4+1]*vect[3];
	out[2] = matrix[0*4+2]*vect[0]+matrix[1*4+2]*vect[1]+matrix[2*4+2]*vect[2]+matrix[3*4+2]*vect[3];
	out[3] = matrix[0*4+3]*vect[0]+matrix[1*4+3]*vect[1]+matrix[2*4+3]*vect[2]+matrix[3*4+3]*vect[3];
}

void Matrix_multVect3f(GLfloat *out, const GLfloat *matrix, const GLfloat *vect, GLfloat w) {
	out[0] = matrix[0*4+0]*vect[0]+matrix[1*4+0]*vect[1]+matrix[2*4+0]*vect[2]+matrix[3*4+0]*w;
	out[1] = matrix[0*4+1]*vect[0]+matrix[1*4+1]*vect[1]+matrix[2*4+1]*vect[2]+matrix[3*4+1]*w;
	out[2] = matrix[0*4+2]*vect[0]+matrix[1*4+2]*vect[1]+matrix[2*4+2]*vect[2]+matrix[3*4+2]*w;
//	out[3] = matrix[0*4+3]*vect[0]+matrix[1*4+3]*vect[1]+matrix[2*4+3]*vect[2]+matrix[3*4+3]*w;
}


