/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#import <Foundation/Foundation.h>
#import <AppKit/AppKit.h>
#include <SDL.h>
#include <vector>
#include "algorithms.h"

using std::vector;

#import "GuiController.h"

@interface InitValues : NSObject {
@public
	vector<const char*> *modes;
	bool nofullscreen;
}
@end

@implementation InitValues
@end

void addAlgorithms(NSMatrix* matrix, struct algorithm *a) {
	float oldHeight = NSHeight([matrix frame]);
	for (int i = [matrix numberOfRows]-1; i >= 0; i--)
		[matrix removeRow: i];

	while (a->func!=NULL) {
		[matrix addRow];
		id cell = [matrix cellAtRow: [matrix numberOfRows]-1 column: 0];
		[cell setTitle: [NSString stringWithCString: a->description]];
		a++;
	}
//	[matrix sizeToCells];
	[matrix sizeToFit];
	float newHeight = NSHeight([matrix frame]);
	NSRect newFrame = NSOffsetRect([matrix frame], 0, -(newHeight-oldHeight));
	[matrix setFrame: newFrame];
}

void addOrders(NSMatrix* matrix) {
	float oldHeight = NSHeight([matrix frame]);
	for (int i = [matrix numberOfRows]-1; i >= 0; i--)
		[matrix removeRow: i];

	struct OrderDesc *od = orders;
	while (od->name!=NULL) {
		[matrix addRow];
		id cell = [matrix cellAtRow: [matrix numberOfRows]-1 column: 0];
		[cell setTitle: [NSString stringWithCString: od->description]];
		od++;
	}
//	[matrix sizeToCells];
	[matrix sizeToFit];
	float newHeight = NSHeight([matrix frame]);
	NSRect newFrame = NSOffsetRect([matrix frame], 0, -(newHeight-oldHeight));
	[matrix setFrame: newFrame];
}

NSSize resizeWindow(NSSize size, id box, float resizeX, float resizeY) {
	NSRect boxFrame = [box frame];
	if (resizeX>0) {
		float resizable = boxFrame.size.width;
		if ([box autoresizingMask] & NSViewMinXMargin)
			resizable += boxFrame.origin.x;
		if ([box autoresizingMask] & NSViewMaxXMargin)
			resizable += size.width - (boxFrame.origin.x+boxFrame.size.width);
		float additionX = resizeX*resizable/boxFrame.size.width;
		size.width += additionX;
	}
	if (resizeY>0) {
		float resizable = boxFrame.size.height;
		if ([box autoresizingMask] & NSViewMinYMargin)
			resizable += boxFrame.origin.y;
		if ([box autoresizingMask] & NSViewMaxYMargin)
			resizable += size.height - (boxFrame.origin.y+boxFrame.size.height);
		float additionY = resizeY*resizable/boxFrame.size.height;
		size.height += additionY;
	}
	return size;
}

void resizeTo(id view, id end) {
#define lowerRightCorner(rect) NSMakePoint((rect).origin.x + (rect).size.width, (rect).origin.y)
	NSPoint viewCorner = lowerRightCorner([view frame]);
	NSPoint endCorner = lowerRightCorner([end frame]);
	float resizeX = viewCorner.x - endCorner.x;
	float resizeY = endCorner.y - viewCorner.y;

	id box = [[view superview] superview];
	id window = [view window];
//#define rect2str(x) [NSStringFromRect(x) cString]

	NSSize size = [[window contentView] frame].size;
	size = resizeWindow(size, box, resizeX, resizeY);

	[window setContentSize: size];

	[end removeFromSuperview];
}

@implementation GuiController

- (void) terminate: (id) sender {
	exit(0);
}

- (IBAction) cancel: (id) sender {
	[self terminate: sender];
}

- (IBAction) start: (id) sender {
	started = YES;
	[NSApp stop: nil];
}

- (IBAction) help: (id) sender {
	NSAlert *alert = [[NSAlert alloc] init];
	[alert setAlertStyle: NSInformationalAlertStyle];
	[alert addButtonWithTitle: @"OK"];
	[alert setMessageText: @"Help"];
	[alert setInformativeText: @
		"Available keys:\n"
		"F1 - switch between fullscreen and window\n"
		"left arrow - slow motion\n"
		"right arrow - fast forward\n"
		"space bar - pause\n"
		"escape - exit"];
//	[alert runModal];
	[alert beginSheetModalForWindow: window modalDelegate: nil didEndSelector: nil contextInfo: NULL];
	[alert release];
}

- (IBAction) updateSelection: (id) sender {
	int n = 0;
	NSMatrix *sort = (NSMatrix*) sortMatrix;
	for (int i = 0; i < [sort numberOfRows]; i++)
		if ([[sort cellAtRow: i column: 0] state]==NSOnState)
			n++;
	NSMatrix *tree = (NSMatrix*) treeMatrix;
	for (int i = 0; i < [tree numberOfRows]; i++)
		if ([[tree cellAtRow: i column: 0] state]==NSOnState)
			n++;
	[ok setEnabled: n > 0 ? YES : NO];
}

- (void) windowWillClose: (NSNotification*) notification {
	if (!started)
		[self terminate: nil];
}

- (void) windowDidExpose: (NSNotification*) notification {
}

- (void) awakeFromNib {
	started = NO;

	InitValues *values = (InitValues*) owner;
	if (values->nofullscreen)
		[fullscreen setEnabled: NO];

	NSPopUpButton *pub = (NSPopUpButton*) resolution;
	[pub removeAllItems];

	int activemode = 0;
	for (int i = 0; i < int(values->modes->size()); i++) {
		const char *mode = values->modes->at(i);
		if (!strcmp(mode, "640x480")) activemode = i;
		[pub addItemWithTitle: [NSString stringWithCString: mode]];
	}
	[pub selectItemAtIndex: activemode];


	[[[sortMatrix superview] superview] setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];
	[[[treeMatrix superview] superview] setAutoresizingMask: NSViewMinXMargin | NSViewHeightSizable];
	[[[orderMatrix superview] superview] setAutoresizingMask: NSViewMinXMargin | NSViewHeightSizable];
	[[[resolution superview] superview] setAutoresizingMask: NSViewMinXMargin | NSViewMaxYMargin];
	addAlgorithms((NSMatrix*) sortMatrix, sortalgorithms);
	resizeTo(sortMatrix, sortEndPos);

	[[[sortMatrix superview] superview] setAutoresizingMask: NSViewMaxXMargin | NSViewHeightSizable];
	[[[treeMatrix superview] superview] setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];
	addAlgorithms((NSMatrix*) treeMatrix, treealgorithms);
	resizeTo(treeMatrix, treeEndPos);

	[[[treeMatrix superview] superview] setAutoresizingMask: NSViewMaxXMargin | NSViewHeightSizable];
	[[[orderMatrix superview] superview] setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];
	[[[resolution superview] superview] setAutoresizingMask: NSViewWidthSizable | NSViewMaxYMargin];
	addOrders((NSMatrix*) orderMatrix);
	resizeTo(orderMatrix, orderEndPos);
	[orderMatrix selectCellAtRow: 0 column: 0];


	[self updateSelection: nil];
	[window center];
	[window makeKeyAndOrderFront: nil];
}
@end

void showGui(vector<threadFunc> &funcs, vector<Anim::AnimType> &types, DataOrder &order, vector<const char*> &modes, int &mode, bool &fullscreen, bool nofullscreen, int &sortelements, int &treeelements) {
	InitValues *values = [[InitValues alloc] init];
	values->modes = &modes;
	values->nofullscreen = nofullscreen;

	NSBundle *bundle = [NSBundle mainBundle];
	NSNib *nib = [[NSNib alloc] initWithNibNamed: @"dst" bundle: bundle];
	NSArray *array;
	[nib instantiateNibWithOwner: values topLevelObjects: &array];

	NSWindow *window = nil;
	GuiController *controller = nil;
	int i;
	for (i = 0; i < int([array count]); i++) 
		if ([[array objectAtIndex: i] isKindOfClass: [NSWindow class]])
			window = [array objectAtIndex: i];
		else if ([[array objectAtIndex: i] isKindOfClass: [GuiController class]])
			controller = [array objectAtIndex: i];

	// modify app menu, set target for Quit to controller
	NSMenu *mainMenu = [NSApp mainMenu];
	NSMenu *appMenu = [[[mainMenu itemArray] objectAtIndex: 0] submenu];
	NSArray *menuArray = [appMenu itemArray];
	id <NSMenuItem> quitItem = [menuArray objectAtIndex: [menuArray count]-1];

	id oldTarget = [quitItem target];
	SEL oldAction = [quitItem action];
	[quitItem setTarget: controller];
	[quitItem setAction: @selector(terminate:)];
	[quitItem setRepresentedObject: window];

//	[NSApp runModalForWindow: window];
	unsetenv("SDL_ENABLEAPPEVENTS");
	[NSApp run];
	setenv("SDL_ENABLEAPPEVENTS", "1", 1);

	struct algorithm *a = sortalgorithms;
	i = 0;
	while (a->func!=NULL) {
		if ([[controller->sortMatrix cellAtRow: i column: 0] state]==NSOnState) {
			funcs.push_back(a->func);
			types.push_back(Anim::GRAPH);
		}
		i++;
		a++;
	}
	a = treealgorithms;
	i = 0;
	while (a->func!=NULL) {
		if ([[controller->treeMatrix cellAtRow: i column: 0] state]==NSOnState) {
			funcs.push_back(a->func);
			types.push_back(Anim::TREE);
		}
		i++;
		a++;
	}

	int orderNr = [controller->orderMatrix selectedRow];
	struct OrderDesc *od = orders;
	for (i = 0; i < orderNr; i++)
		od++;
	order = od->order;

	mode = [controller->resolution indexOfSelectedItem];
	fullscreen = [controller->fullscreen state]==NSOnState;
	sortelements = [controller->sortElements intValue];
	treeelements = [controller->treeElements intValue];

/*
	int time = SDL_GetTicks();
	while (SDL_GetTicks()-time<5000) {
		SDL_Event event;
		SDL_WaitEvent(&event);
	}
*/

	[window setExcludedFromWindowsMenu: YES];
	[NSApp removeWindowsItem: window];

	[window close];

	[nib release];
	[bundle release];

	[quitItem setTarget: oldTarget];
	[quitItem setAction: oldAction];

	[values release];
}

