/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifdef HAVE_SDL_TTF
#include <SDL.h>
#include <SDL_ttf.h>
#include <SDL_opengl.h>
#include <stdarg.h>
#include <string.h>
#include "sdlttf.h"

#define TEXWIDTH 128
#define TEXHEIGHT 128




Font::Font(const char *file, int size, int first, int last) {
	bool init = !TTF_WasInit();
	if (init) TTF_Init();

	TTF_Font *font = TTF_OpenFont(file,size);
	if (font==NULL) return;

	offset = first;
	max = last;
//	lists = glGenLists(max-offset+1);
	lists = glGenLists(max-offset+1 + 35);

	numtextures = 0;
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	tmpsurf = SDL_CreateRGBSurface(SDL_SWSURFACE,TEXWIDTH,TEXHEIGHT,32,0x000000FF,0x0000FF00,0x00FF0000,0xFF000000);
#else
	tmpsurf = SDL_CreateRGBSurface(SDL_SWSURFACE,TEXWIDTH,TEXHEIGHT,32,0xFF000000,0x00FF0000,0x0000FF00,0x000000FF);
#endif
	initSurf();

	for (int i = offset; i <= max; i++)
		addGlyph(font,i);

	uploadSurf();
	SDL_FreeSurface(tmpsurf);

	if (offset > 'A' || max < 'A') {
		SDL_Color color = { 1, 1, 1, 0 };
		SDL_Surface *surf = TTF_RenderGlyph_Blended(font, 'A', color);
		heightA = surf->h;
		SDL_FreeSurface(surf);
	} else {
		heightA = height['A'-offset];
	}

	TTF_CloseFont(font);

	if (init) TTF_Quit();
}

Font::~Font() {
	glDeleteLists(lists,max-offset+1);
	glDeleteTextures(numtextures,textures);
}

void Font::initSurf() {
	SDL_FillRect(tmpsurf,NULL,0);
	glGenTextures(1,&textures[numtextures]);
	glBindTexture(GL_TEXTURE_2D,textures[numtextures]);
	numtextures++;
	curx = cury = nexty = 0;
}

void Font::uploadSurf() {
	if (SDL_MUSTLOCK(tmpsurf)) SDL_LockSurface(tmpsurf);
	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA, TEXWIDTH, TEXHEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, tmpsurf->pixels);
	if (SDL_MUSTLOCK(tmpsurf)) SDL_UnlockSurface(tmpsurf);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
}

void Font::newTexture() {
	uploadSurf();
	initSurf();
}

void Font::addGlyph(TTF_Font *font, int c) {
	int margin = 1;
	SDL_Color color = { 1, 1, 1, 0 };
	SDL_Surface *surf = TTF_RenderGlyph_Blended(font,c,color);
	if ((curx+surf->w+margin>TEXWIDTH && nexty+surf->h+margin>TEXHEIGHT) || cury+surf->h+margin>TEXHEIGHT) {
		newTexture();
	}
	if (curx+surf->w+margin>TEXWIDTH) {
		curx = 0;
		cury = nexty;
	}

	int minx, maxx, miny, maxy, adv;
	TTF_GlyphMetrics(font,c,&minx,&maxx,&miny,&maxy,&adv);
	advance[c-offset] = adv;
	height[c-offset] = surf->h;
	float x1 = float(curx)/TEXWIDTH;
	float x2 = float(curx+surf->w)/TEXWIDTH;
	float y1 = float(cury)/TEXHEIGHT;
	float y2 = float(cury+surf->h)/TEXHEIGHT;

	SDL_SetAlpha(surf,0,255);
	SDL_Rect dst;
	dst.x = curx;
	dst.y = cury;
	SDL_BlitSurface(surf,NULL,tmpsurf,&dst);

	if (cury+surf->h+margin>nexty) nexty = cury+surf->h+margin;
	curx += surf->w+margin;

	glNewList(lists+c-offset,GL_COMPILE);
	glBindTexture(GL_TEXTURE_2D,textures[numtextures-1]);
	glBegin(GL_QUADS);
	glTexCoord2f(x1,y2);
	glVertex2f(minx,miny);
	glTexCoord2f(x2,y2);
	glVertex2f(minx+surf->w,miny);
	glTexCoord2f(x2,y1);
	glVertex2f(minx+surf->w,miny+surf->h);
	glTexCoord2f(x1,y1);
	glVertex2f(minx,miny+surf->h);
	glEnd();
	glTranslatef(adv,0,0);
	glEndList();

	SDL_FreeSurface(surf);
}

void Font::print(float height, const char *fmt, ...) {
	char buffer[1024];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buffer,fmt,ap);
	va_end(ap);

	int iw, ih;
	getSize(buffer,&iw,&ih);

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glPushMatrix();
	glScalef(height/ih,height/ih,0);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	glListBase(lists-offset);
	glCallLists(strlen(buffer),GL_UNSIGNED_BYTE,buffer);

	glPopMatrix();
	glPopAttrib();
}

void Font::printCentered(float height, const char *fmt, ...) {
	char buffer[1024];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buffer,fmt,ap);
	va_end(ap);

	int iw, ih;
	getSize(buffer,&iw,&ih);
	float fw = iw*height/ih;

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glPushMatrix();
	glTranslatef(-fw/2,-height/2,0);
	glScalef(height/ih,height/ih,0);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	glListBase(lists-offset);
	glCallLists(strlen(buffer),GL_UNSIGNED_BYTE,buffer);

	glPopMatrix();
	glPopAttrib();
}

void Font::getSize(const char *str, int *wptr, int *hptr) {
	*wptr = 0;
	*hptr = heightA;
	for (; *str; str++) {
		if (*str<offset) continue;
		*wptr += advance[*str-offset];
	}
}

void Font::printCentered2D(float height, const char *fmt, ...) {
	char buffer[1024];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buffer,fmt,ap);
	va_end(ap);

	GLdouble model[16], proj[16];
	GLint view[4];
	glGetDoublev(GL_MODELVIEW_MATRIX,model);
	glGetDoublev(GL_PROJECTION_MATRIX,proj);
	glGetIntegerv(GL_VIEWPORT,view);

	GLdouble winX1, winY1, winZ1, winX2, winY2, winZ2;
	gluProject(0,height/2,0,model,proj,view,&winX1,&winY1,&winZ1);
	gluProject(0,-height/2,0,model,proj,view,&winX2,&winY2,&winZ2);

	height = winY1-winY2;

	int iw, ih;
	getSize(buffer,&iw,&ih);
	float fw = iw*height/ih;

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(view[0],view[0]+view[2],view[1],view[1]+view[3],-1,1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(winX1-fw/2,winY2,0);
	glScalef(height/ih,height/ih,0);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	glListBase(lists-offset);
	glCallLists(strlen(buffer),GL_UNSIGNED_BYTE,buffer);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPopAttrib();
}

void Font::printOverlay2D(float height, float x, float y, const char *fmt, ...) {
	char buffer[1024];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buffer,fmt,ap);
	va_end(ap);

	GLint view[4];
	glGetIntegerv(GL_VIEWPORT,view);

	drawOverlay2D(height*view[3], x*view[2], y*view[3], buffer);
}

void Font::printOverlay2D(int height, int x, int y, const char *fmt, ...) {
	char buffer[1024];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buffer,fmt,ap);
	va_end(ap);

	drawOverlay2D(height, x, y, buffer);
}

void Font::drawOverlay2D(float height, float x, float y, char *buffer) {

	GLint view[4];
	glGetIntegerv(GL_VIEWPORT,view);

	int iw, ih;
	getSize(buffer,&iw,&ih);

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(view[0],view[0]+view[2],view[1],view[1]+view[3],-1,1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(view[0]+x,view[1]+y-height,0);
	glScalef(height/ih,height/ih,0);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	glListBase(lists-offset);
	glCallLists(strlen(buffer),GL_UNSIGNED_BYTE,buffer);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPopAttrib();
}

#endif
