/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <SDL_opengl.h>
#include "sphere.h"
#include "vect.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif


void Sphere::makeRing(float pos, GLfloat *ring) {
	float t = -1 + 2*pos;
	float r = sqrt(1-t*t);
	for (int j = 0; j < n; j++) {
		float angle = 2*M_PI*float(j)/n;
		float c = cos(angle), s = sin(angle);
		GLfloat *point = &ring[3*j];
		point[0] = r*c;
		point[1] = r*s;
		point[2] = t;
	}
}

void Sphere::makeArc(float pos, GLfloat *arc) {
	float angle = 2*M_PI*pos;
	float c = cos(angle), s = sin(angle);
	for (int i = 0; i < m; i++) {
		float t = -1 + 2*float(i)/(m-1);
		float r = sqrt(1-t*t);
		GLfloat *point = &arc[3*i];
		point[0] = r*c;
		point[1] = r*s;
		point[2] = t;
	}
}

void Sphere::drawRing(float radius, GLfloat *ring1, GLfloat *ring2) {
	for (int j = 0; j < n; j++) {
		int i1 = 3*j;
		int i2 = 3*((j+1)%n);
		int i3 = 3*((j+1)%n);
		int i4 = 3*j;
		GLfloat p1[3], p2[3], p3[3], p4[3];
		Vect_copy3f(p1,&ring1[i1]);
		Vect_mult3f(p1,radius);
		Vect_copy3f(p2,&ring1[i2]);
		Vect_mult3f(p2,radius);
		Vect_copy3f(p3,&ring2[i3]);
		Vect_mult3f(p3,radius);
		Vect_copy3f(p4,&ring2[i4]);
		Vect_mult3f(p4,radius);
		glNormal3fv(&ring1[i1]);
		glVertex3fv(p1);
		glNormal3fv(&ring1[i2]);
		glVertex3fv(p2);
		glNormal3fv(&ring2[i3]);
		glVertex3fv(p3);
		glNormal3fv(&ring2[i4]);
		glVertex3fv(p4);
	}
}

void Sphere::drawArc(float radius, GLfloat *arc1, int pitch1, GLfloat *arc2, int pitch2) {
	for (int i = 0; i < m-1; i++) {
		int i1 = pitch1*i;
		int i2 = pitch2*i;
		int i3 = pitch2*(i+1);
		int i4 = pitch1*(i+1);
		GLfloat p1[3], p2[3], p3[3], p4[3];
		Vect_copy3f(p1,&arc1[i1]);
		Vect_mult3f(p1,radius);
		Vect_copy3f(p2,&arc2[i2]);
		Vect_mult3f(p2,radius);
		Vect_copy3f(p3,&arc2[i3]);
		Vect_mult3f(p3,radius);
		Vect_copy3f(p4,&arc1[i4]);
		Vect_mult3f(p4,radius);
		glNormal3fv(&arc1[i1]);
		glVertex3fv(p1);
		glNormal3fv(&arc2[i2]);
		glVertex3fv(p2);
		glNormal3fv(&arc2[i3]);
		glVertex3fv(p3);
		glNormal3fv(&arc1[i4]);
		glVertex3fv(p4);
	}
}

Sphere::Sphere(int n, int m) {
	this->n = n;
	this->m = m;
	points = new GLfloat[3*m*n];
	for (int i = 0; i < m; i++) {
		makeRing(float(i)/(m-1),&points[3*n*i]);
	}
}

Sphere::~Sphere() {
	delete [] points;
}

void Sphere::draw(float radius, float start, float end) {
	GLfloat *startring = NULL;
	GLfloat *endring = NULL;
	int s = 0, e = m-1;
	if (fabs(start)>0.01) {
		startring = new GLfloat[3*n];
		s = int(ceil(start*(m-1)));
		makeRing(start,startring);
	}
	if (fabs(end-1)>0.01) {
		endring = new GLfloat[3*n];
		e = int(floor(end*(m-1)));
		makeRing(end,endring);
	}
	glBegin(GL_QUADS);
	if (startring) drawRing(radius,startring,&points[3*n*s]);
	for (int i = s; i < e; i++)
		drawRing(radius,&points[3*n*i],&points[3*n*(i+1)]);
	if (endring) drawRing(radius,&points[3*n*e],endring);
	glEnd();
	if (startring) delete [] startring;
	if (endring) delete [] endring;
}

void Sphere::drawSector(float radius, float start, float end) {
	GLfloat *startarc = NULL;
	GLfloat *endarc = NULL;
	int s = 0, e = n;
	if (fabs(start)>0.01) {
		startarc = new GLfloat[3*m];
		s = int(ceil(start*n));
		makeArc(start,startarc);
	}
	if (fabs(end-1)>0.01) {
		endarc = new GLfloat[3*m];
		e = int(floor(end*n));
		makeArc(end,endarc);
	}
	glBegin(GL_QUADS);
	if (startarc) drawArc(radius,startarc,3,&points[3*(s%n)],3*n);
	for (int i = s; i < e; i++)
		drawArc(radius,&points[3*i],3*n,&points[3*((i+1)%n)],3*n);
	if (endarc) drawArc(radius,&points[3*(e%n)],3*n,endarc,3);
	glEnd();
	if (startarc) delete [] startarc;
	if (endarc) delete [] endarc;
}


