/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <SDL_opengl.h>
#include "spline.h"
#include "vect.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

void makeWaypoint(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, GLfloat *waypoint, float fact = 0.2) {
	GLfloat tempV1[3], tempV2[3];
	// tempV1 = v1-v0
	Vect_copy3f(tempV1,v1);
	Vect_sub3f(tempV1,v0);
	// tempV2 = v2-v1
	Vect_copy3f(tempV2,v2);
	Vect_sub3f(tempV2,v1);
	float l1 = Vect_abs3f(tempV1);
	float l2 = Vect_abs3f(tempV2);
	if (l1<l2)
		Vect_mult3f(tempV2,l1/l2);
	else if (l1>l2)
		Vect_mult3f(tempV1,l2/l1);
	Vect_copy3f(waypoint,tempV1);
	Vect_add3f(waypoint,tempV2);
	Vect_mult3f(waypoint,fact);
	Vect_add3f(waypoint,v1);
}

Spline::Spline(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3) {
	init(v0,v1,v2,v3);
}

Spline::Spline() {
}

Spline::~Spline() {
}

void Spline::init(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3) {
	if (v0==NULL) v0 = v1;
	if (v3==NULL) v3 = v2;
	Vect_copy3f(start,v1);
	Vect_copy3f(end,v2);

	makeWaypoint(v0,v1,v2,waypoint1);
	makeWaypoint(v3,v2,v1,waypoint3);

	Vect_copy3f(waypoint2,waypoint1);
	Vect_add3f(waypoint2,waypoint3);
	Vect_mult3f(waypoint2,0.5);

	float l1 = getLength(start,waypoint1,waypoint2,&len1,&len2);
	float l2 = getLength(waypoint2,waypoint3,end,&len3,&len4);
	if (l1+l2==0) divpos = 0;
	else divpos = l1/(l1+l2);
	totlen = l1+l2;
}

float Spline::getLength(const GLfloat *v1, const GLfloat *v2, const GLfloat *v3, float *lp1, float *lp2) {
	GLfloat tempV1[3], tempV2[3];
	Vect_copy3f(tempV1,v2);
	Vect_sub3f(tempV1,v1);
	Vect_copy3f(tempV2,v3);
	Vect_sub3f(tempV2,v2);
	float l1 = Vect_abs3f(tempV1);
	float l2 = Vect_abs3f(tempV2);
	*lp1 = l1;
	*lp2 = l2;
	if (l2>l1) {
		float tmp = l2;
		l2 = l1;
		l1 = tmp;
	}
	float dot = Vect_dot3f(tempV1,tempV2);
	if (l1>0) dot /= l1;
	if (l2>0) dot /= l2;
	return l1 + l2*(0.5+0.5*dot);
}

float Spline::getLength() {
	return totlen;
}

void Spline::calc(float pos, GLfloat *out) {
	GLfloat *v1 = start;
	GLfloat *v2 = waypoint1;
	GLfloat *v3 = waypoint2;
	float l1 = len1;
	float l2 = len2;
	if (pos<divpos) {
		pos = pos/divpos;
	} else {
		v1 = waypoint2;
		v2 = waypoint3;
		v3 = end;
		l1 = len3;
		l2 = len4;
		if (divpos<1)
			pos = (pos-divpos)/(1-divpos);
	}
	if (l1!=l2)
		pos = (l1-sqrt(l1*l1+(l2-l1)*((l1+l2)*pos)))/(l1-l2);
	GLfloat tempV[3];
	Vect_copy3f(out,v1);
	Vect_mult3f(out,1-2*pos+pos*pos);
	Vect_copy3f(tempV,v2);
	Vect_mult3f(tempV,2*pos-2*pos*pos);
	Vect_add3f(out,tempV);
	Vect_copy3f(tempV,v3);
	Vect_mult3f(tempV,pos*pos);
	Vect_add3f(out,tempV);
}

Spline3::Spline3(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3) {
	init(v0,v1,v2,v3);
}

Spline3::Spline3() {
}

Spline3::~Spline3() {
}

void Spline3::init(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3) {
	if (v0==NULL) v0 = v1;
	if (v3==NULL) v3 = v2;
	Vect_copy3f(start,v1);
	Vect_copy3f(end,v2);

	makeWaypoint(v0,v1,v2,waypoint1);
	makeWaypoint(v3,v2,v1,waypoint2);

	length = -1;
}

float Spline3::getLength() {
	if (length>=0) return length;
#define INTEGRALSTEPS 10
/*
#define INTEGRALDIFF (float(1)/(3*INTEGRALSTEPS))
	length = 0;
	for (int i = 0; i < INTEGRALSTEPS; i++) {
		GLfloat pos1[3], pos2[3];
		calc(float(i)/INTEGRALSTEPS,pos1);
		calc(float(i)/INTEGRALSTEPS+INTEGRALDIFF,pos2);
		float xdiff = (pos2[0]-pos1[0])/INTEGRALDIFF;
		float ydiff = (pos2[1]-pos1[1])/INTEGRALDIFF;
		float zdiff = (pos2[2]-pos1[2])/INTEGRALDIFF;
		length += hypot(xdiff,hypot(ydiff,zdiff));
	}
	length /= INTEGRALSTEPS;
*/
	length = 0;
	GLfloat oldpos[3];
	calc(0,oldpos);
	for (int i = 1; i <= INTEGRALSTEPS; i++) {
		GLfloat pos[3];
		calc(float(i)/INTEGRALSTEPS,pos);
		float xdiff = (pos[0]-oldpos[0]);
		float ydiff = (pos[1]-oldpos[1]);
		float zdiff = (pos[2]-oldpos[2]);
		length += hypot(xdiff,hypot(ydiff,zdiff));
		Vect_copy3f(oldpos,pos);
	}
	return length;
}

void Spline3::calc(float pos, GLfloat *out) {
	// r = start*(1-t)^3 + waypoint1*t*(1-t)^2 + waypoint2*(1-t)*t^2 + end*t^3
	// According to Kreyszig, Advanced Engineering Mathematics 8th ed., page 868:
	// r = r0 + v0*t + (3*(r1-r0)-(2*v0+v1))*t^2 + (2*(r0-r1)+v0+v1)*t^3
	// r0 = start
	// r1 = end
	// waypoint1 = r0 + v0/3
	// waypoint2 = r1 - v1/3
	// r(0) = r0, r(1) = r1, r'(0) = v0, r'(1) = v1
	GLfloat *v1 = start;
	GLfloat *v2 = waypoint1;
	GLfloat *v3 = waypoint2;
	GLfloat *v4 = end;
	GLfloat tempV[3];
	Vect_copy3f(out,v1);
	Vect_mult3f(out,(1-pos)*(1-pos)*(1-pos));

	Vect_copy3f(tempV,v2);
	Vect_mult3f(tempV,3*(1-pos)*(1-pos)*pos);
	Vect_add3f(out,tempV);

	Vect_copy3f(tempV,v3);
	Vect_mult3f(tempV,3*(1-pos)*pos*pos);
	Vect_add3f(out,tempV);

	Vect_copy3f(tempV,v4);
	Vect_mult3f(tempV,pos*pos*pos);
	Vect_add3f(out,tempV);
}


Rotation::Rotation() {
}

Rotation::Rotation(const GLfloat *m0, const GLfloat *m1, const GLfloat *m2, const GLfloat *m3, int a) {
	init(m0,m1,m2,m3,a);
}

Rotation::Rotation(const GLfloat *m1, const GLfloat *m2, int a) {
	init(NULL,m1,m2,NULL,a);
}

Rotation::~Rotation() {
}

void Rotation::init(const GLfloat *m0, const GLfloat *m1, const GLfloat *m2, const GLfloat *m3, int a) {
	if (m0==NULL) m0 = m1;
	if (m3==NULL) m3 = m2;
	posSpline.init(&m0[12],&m1[12],&m2[12],&m3[12]);
	ax1 = a;
	ax2 = (ax1+1)%3;
	ax3 = (ax2+1)%3;
	spline1.init(&m0[4*ax1],&m1[4*ax1],&m2[4*ax1],&m3[4*ax1]);
	spline2.init(&m0[4*ax2],&m1[4*ax2],&m2[4*ax2],&m3[4*ax2]);
}

void Rotation::init(const GLfloat *m1, const GLfloat *m2, int a) {
	init(NULL,m1,m2,NULL,a);
}

void Rotation::rotate(float pos, GLfloat *out) {
	posSpline.calc(pos,&out[12]);
	spline1.calc(pos,&out[4*ax1]);
	Vect_normalize3f(&out[4*ax1]);
	spline2.calc(pos,&out[4*ax2]);
	Vect_cross3f(&out[4*ax3],&out[4*ax1],&out[4*ax2]);
	Vect_cross3f(&out[4*ax2],&out[4*ax3],&out[4*ax1]);
	Vect_normalize3f(&out[4*ax2]);
	Vect_normalize3f(&out[4*ax3]);
	out[15] = 1;
	out[3] = 0;
	out[7] = 0;
	out[11] = 0;
}

