/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL.h>
#include <stdio.h>
#include <math.h>
#include "tree.h"
#include "vect.h"
#include "camera.h"
#include "sphere.h"
#include "matrix.h"
#include "sdlttf.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define NODESTYLE 4

#ifndef HAVE_SDL_TTF
#if NODESTYLE == 4
#undef NODESTYLE
#define NODESTYLE 3
#endif
#endif

Sphere sphere(17,8);
extern Font *treefont;
GLUquadric *quadric = NULL;
int quadricusers = 0;

Node::Node(int v) {
	val = v;
	left = right = NULL;

	parentlength = 1;
	spacing = false;
	defaultColor = true;
	hide = false;
	red = false;
}

Node::Node(float w, float h) {
	val = -1;
	spacing = true;
	width = w;
	height = h;
	left = right = NULL;
	hide = true;
	red = false;
}

Node::Node(int v, float w, float h) {
	val = v;
	spacing = true;
	width = w;
	height = h;
	left = right = NULL;
	hide = true;
	red = false;
}

Node::~Node() {
	delete left;
	delete right;
}

float Node::updateWidth() {
	if (spacing) return width;
	width = NODEWIDTH;
	if (left) width += left->updateWidth();
	if (right) width += right->updateWidth();
	return width;
}

float Node::updateHeight() {
	if (spacing) return height;
	height = 0;
	iheight = 0;
	balance = 0;
	if (left) {
		height = NODEHEIGHT+left->updateHeight();
		iheight = 1+left->iheight;
		balance += (1+left->iheight);
	}
	if (right) {
		float d2 = NODEHEIGHT+right->updateHeight();
		if (d2>height) height = d2;
		int id2 = 1+right->iheight;
		if (id2>iheight) iheight = id2;
		balance -= (1+right->iheight);
	}
	return height;
}

void Node::drawCylinder(GLfloat *end, float radius) {
	float yrot = 180*atan2(end[0],end[2])/M_PI;
	float hdist = hypot(end[0],end[2]);
	float xrot = -180*atan2(end[1],hdist)/M_PI;
	float dist = hypot(end[1],hdist);
	GLfloat matSpecular[] = { 0, 0, 0, 0 };
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecular);
	glPushMatrix();
	glRotatef(yrot,0,1,0);
	glRotatef(xrot,1,0,0);
	if (parentlength<0) {
		parentlength = -parentlength;
		glTranslatef(0,0,dist*(1-parentlength));
	}
	gluCylinder(quadric,radius,radius,dist*parentlength,9,1);
	glPopMatrix();
}


void Node::drawNode() {
	if (hide) return;

	GLfloat c2[3];
	Vect_copy3f(c2,color);
	c2[0] += 1;
	c2[1] += 1;
	c2[2] += 1;
	Vect_mult3f(c2,0.5);

	glColor3f(color[0],color[1],color[2]);
	GLfloat matSpecular[] = { 0.8, 0.8, 0.8, 1 };
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecular);

#if NODESTYLE == 1
	gluSphere(quadric,0.45*NODEWIDTH,20,10);
	float d = NODEWIDTH*0.90;
	drawBox(1,1,1, -d/10,-d*0.30,(1-0.15)*d/2, d/5,d*0.60,0.15*d/2);
	float size = float(val)/maxval;
	drawBox(0,0,0.3, -d/10+d/34,-d*0.30+d/34,(1-0.15)*d/2, d/5-d/17,(d*0.60-d/17)*size,0.18*d/2);
#elif NODESTYLE == 2
	glPushMatrix();
	glRotatef(-90,1,0,0);
	sphere.draw(0.45*NODEWIDTH,0,float(val)/maxval);
	glColor3f(c2[0],c2[1],c2[2]);
	sphere.draw(0.45*NODEWIDTH,float(val)/maxval,1);
	glPopMatrix();
#elif NODESTYLE == 3
	glPushMatrix();
	glRotatef(90,0,0,1);
	sphere.drawSector(0.45*NODEWIDTH,1-float(val)/maxval,1);
	glColor3f(c2[0],c2[1],c2[2]);
	sphere.drawSector(0.45*NODEWIDTH,0,1-float(val)/maxval);
	glPopMatrix();
#elif NODESTYLE == 4
	glPushMatrix();
	sphere.draw(0.45*NODEWIDTH);
//	glTranslatef(0,0,0.451*NODEWIDTH);
	glColor3f(1,1,1);
	treefont->printCentered2D(0.35*NODEWIDTH,"%d",val);
	glPopMatrix();
#endif

}

void Node::draw(GLfloat *parent) {
	if (spacing) {
//		drawBox(0,1,0, 0,0,0, width,-height,0.2);
		return;
	}
	GLfloat pos[] = { 0, 0, 0, 1 };
	if (left) pos[0] += left->width;
	pos[0] += NODEWIDTH/2;
	pos[1] -= NODEWIDTH/2;
	glPushMatrix();
	glTranslatef(pos[0],pos[1],pos[2]);
	if (parent) {
		GLfloat ppos[3];
		Vect_copy3f(ppos,parent);
		Vect_sub3f(ppos,pos);
		glColor3f(1,1,0.5);
		drawCylinder(ppos,0.125*NODEWIDTH);
	}
	drawNode();
	glPopMatrix();
	pos[1] += NODEHEIGHT;
	float leftwidth = 0;
	if (left) {
		leftwidth = left->width;
		glPushMatrix();
		glTranslatef(0,-NODEHEIGHT,0);
		left->draw(pos);
		glPopMatrix();
	}
	if (right) {
		glPushMatrix();
		glTranslatef(leftwidth+NODEWIDTH,-NODEHEIGHT,0);
		pos[0] -= leftwidth+NODEWIDTH;
		right->draw(pos);
		glPopMatrix();
	}
}

void Node::addPos(Node *n, GLfloat *pos) {
	if (spacing) return;
	if (n->val<val) {
		pos[1] -= NODEHEIGHT;
		if (left) left->addPos(n,pos);
	} else if (n->val>val) {
		pos[1] -= NODEHEIGHT;
		if (left) pos[0] += left->width;
		pos[0] += NODEWIDTH;
		if (right) right->addPos(n,pos);
	} else {
		if (left) pos[0] += left->width;
	}
}

void Node::setMaxval(int max) {
	maxval = max;
	if (left) left->setMaxval(max);
	if (right) right->setMaxval(max);
}


void Node::setDefaultColor(GLfloat *c) {
	if (defaultColor) memcpy(color,c,sizeof(GLfloat)*3);
	if (left) left->setDefaultColor(c);
	if (right) right->setDefaultColor(c);
}



Tree::Tree() {
	root = NULL;
	type = TYPE_ABSOLUTE;
}

Tree::~Tree() {
	delete root;
}



Node **Tree::getParentLink(int i) {
	Node *n = root, *parent = NULL;
	int dir = 0;
	if (root==NULL) {
		return &root;
	}
	while (true) {
		if (n==NULL || n->spacing) {
			if (dir<0) return &parent->left;
			else return &parent->right;
		}
		if (n->val>i) {
			parent = n;
			n = n->left;
			dir = -1;
		} else if (n->val<i) {
			parent = n;
			n = n->right;
			dir = 1;
		} else {
			if (parent==NULL) return &root;
			if (dir<0) return &parent->left;
			else return &parent->right;
		}
	}
}

Node **Tree::getHeapLink(int pos) {
	int parent = pos/2;
	if (parent==0) return &root;
	if (pos==2*parent) return &heap[parent]->left;
	return &heap[parent]->right;
}

void Tree::addHeapSubTreePos(int index, GLfloat *pos) {
	int parent = index/2;
	if (parent==0) return;
	if (heap[parent]==NULL) return;
	if (index==2*parent+1)
		pos[0] += heap[parent]->left->width + NODEWIDTH;
	pos[1] -= NODEHEIGHT;
	addHeapSubTreePos(parent,pos);
}

void Tree::addHeapPos(int index, GLfloat *pos) {
	addHeapSubTreePos(index,pos);
	if (index<int(heap.size())) {
		if (heap[index]->left) {
			pos[0] += heap[index]->left->width;
		}
	}
}



void Tree::add(int i, Node *newnode) {
	Node **link = getParentLink(i);
	if (*link==NULL) *link = newnode;
}

void Tree::add(int i) {
	add(i,new Node(i));
}

void Tree::addPos(Node *n, GLfloat *pos) {
	if (root!=NULL) root->addPos(n,pos);
}




static void initQuadrics() {
	quadric = gluNewQuadric();
	gluQuadricNormals(quadric,GLU_SMOOTH);
	gluQuadricTexture(quadric,GL_FALSE);
	gluQuadricOrientation(quadric,GLU_OUTSIDE);
	gluQuadricDrawStyle(quadric,GLU_FILL);
}

static void deleteQuadrics() {
	gluDeleteQuadric(quadric);
}

void reinitQuadrics() {
	if (quadricusers>0) {
		deleteQuadrics();
		initQuadrics();
	}
}

Forest::Forest() {
//	GLfloat black[] = { 0.15, 0.15, 0.15 };
	GLfloat blue[] = { 0, 0, 1 };
	memcpy(defaultColor,blue,sizeof(GLfloat)*3);
	if (quadricusers==0)
		initQuadrics();
	quadricusers++;
}

Forest::~Forest() {
	quadricusers--;
	if (quadricusers==0) deleteQuadrics();
	clear();
}
void Forest::clear() {
	int n = trees.size();
	for (int i = 0; i < n; i++)
		delete trees[i];
	trees.clear();
	n = movables.size();
	for (int i = 0; i < n; i++)
		delete movables[i];
	movables.clear();
}

void Forest::getDefaultColor(GLfloat *color) {
	memcpy(color,defaultColor,sizeof(GLfloat)*3);
}

void Forest::setMinSize(GLfloat *upleft, GLfloat *upright, GLfloat *downleft) {
	GLfloat width[3], height[3];
	Vect_copy3f(width,upright);
	Vect_sub3f(width,upleft);
	Vect_copy3f(height,upleft);
	Vect_sub3f(height,downleft);
	float w = Vect_abs3f(width);
	float h = Vect_abs3f(height);
	Vect_normalize3f(width);
	Vect_normalize3f(height);
	float minwidth = 5*NODEWIDTH;
	float minheight = 4*NODEHEIGHT;
	if (w<minwidth) {
		float extra = minwidth-w;
		GLfloat addition[3];
		Vect_copy3f(addition,width);
		Vect_mult3f(addition,extra/2);
		Vect_add3f(upright,addition);
		Vect_sub3f(upleft,addition);
		Vect_sub3f(downleft,addition);
	}
	if (h<minheight) {
		float extra = minheight-h;
		GLfloat addition[3];
		Vect_copy3f(addition,height);
		Vect_mult3f(addition,extra/2);
		Vect_add3f(upleft,addition);
		Vect_add3f(upright,addition);
		Vect_sub3f(downleft,addition);
	}
}

void Forest::draw(int time, int difftime) {

	GLfloat lightSpecular[] = { 0.2, 0.2, 0.2, 1 };
	GLfloat lightDiffuse[] = { 0.4, 0.4, 0.4, 1 };
	GLfloat lightAmbient[] = { 0.1, 0.1, 0.1, 1 };
//	GLfloat light0Pos[] = { 0, 5, 10, 1 };
//	GLfloat light1Pos[] = { 20, 5, 10, 1 };
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glEnable(GL_COLOR_MATERIAL);
	glLightfv(GL_LIGHT0, GL_SPECULAR, lightSpecular);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
	glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
	glLightfv(GL_LIGHT1, GL_SPECULAR, lightSpecular);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightDiffuse);
	glLightfv(GL_LIGHT1, GL_AMBIENT, lightAmbient);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 5);


	glGetFloatv(GL_MODELVIEW_MATRIX,modelview);
	glGetIntegerv(GL_VIEWPORT,viewport);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	gluPerspective(45,float(viewport[2])/viewport[3],0.1,100);


	GLfloat upleft[4], upright[4], downleft[4];
	calcView(upleft,upright,downleft);
	glMatrixMode(GL_PROJECTION);
	setCamera(upleft,upright,downleft,45,true,NULL);
//	drawBox(0,1,0, upleft[0],upleft[1],-0.5, upright[0]-upleft[0],downleft[1]-upleft[1],0.2);

	glMatrixMode(GL_MODELVIEW);

	upleft[2] += 10;
	upright[2] += 10;
	glLightfv(GL_LIGHT0, GL_POSITION, upleft);
	glLightfv(GL_LIGHT1, GL_POSITION, upright);


	glPushMatrix();
	int n = trees.size();
	for (int i = 0; i < n; i++) {
		if (trees[i]->root!=NULL) {
			trees[i]->root->updateWidth();
			trees[i]->root->setDefaultColor(defaultColor);
			trees[i]->root->setMaxval(maxval);
			trees[i]->root->draw(NULL);
			glTranslatef(trees[i]->root->width,0,0);
		}
		glTranslatef(TREESPACING,0,0);
	}
	glPopMatrix();

	n = movables.size();
	for (int i = 0; i < n; i++) {
		Tree *t = movables[i];
		if (t->root==NULL) continue;
		t->root->updateWidth();
		t->root->setDefaultColor(defaultColor);
		glPushMatrix();
		glTranslatef(t->pos[0],t->pos[1],t->pos[2]);
		if (t->type==Tree::TYPE_RELATIVE) {
			GLfloat pos[3];
			getTreePos(t->reltree,pos);
			trees[t->reltree]->addPos(t->relnode,pos);
			glTranslatef(pos[0],pos[1],pos[2]);
		}
		t->root->setMaxval(maxval);
		t->root->draw(NULL);
		glPopMatrix();
	}


	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

}


void Forest::getTreePos(int tree, GLfloat *pos) {
	pos[0] = 0;
	pos[1] = 0;
	pos[2] = 0;
	for (int i = 0; i < tree; i++) {
		if (trees[i]->root!=NULL) {
			trees[i]->root->updateWidth();
			trees[i]->root->updateHeight();
			pos[0] += trees[i]->root->width;
		}
		pos[0] += TREESPACING;
	}
}

void Forest::getTreePos(Tree *t, GLfloat *pos) {
	pos[0] = 0;
	pos[1] = 0;
	pos[2] = 0;
	int ntrees = trees.size();
	for (int i = 0; i < ntrees; i++) {
		if (trees[i]==t) return;
		if (trees[i]->root!=NULL) {
			trees[i]->root->updateWidth();
			trees[i]->root->updateHeight();
			pos[0] += trees[i]->root->width;
		}
		pos[0] += TREESPACING;
	}
	int nmovables = movables.size();
	for (int i = 0; i < nmovables; i++) {
		if (movables[i]==t) {
			memcpy(pos,t->pos,sizeof(GLfloat)*3);
			return;
		}
	}
}

void Forest::calcView(GLfloat *upleft, GLfloat *upright, GLfloat *downleft) {
	float width = 0;
	float height = 0;
	int n = trees.size();
	for (int i = 0; i < n; i++) {
		if (trees[i]->root==NULL) continue;
		trees[i]->root->updateWidth();
		trees[i]->root->updateHeight();
		width += trees[i]->root->width;
		float h = trees[i]->root->height;
		if (h>height) height = h;
	}

	GLfloat ul[] = {0, 0, 1, 1 } , ur[] = { 0, 0, 1, 1 }, dl[] = { 0, 0, 1, 1 };
	memcpy(upleft,ul,sizeof(GLfloat)*4);
	memcpy(upright,ur,sizeof(GLfloat)*4);
	memcpy(downleft,dl,sizeof(GLfloat)*4);
	downleft[1] -= height+NODEWIDTH;
	upright[0] += width + TREESPACING*(n-1);
	upleft[1] += NODEWIDTH;
	upright[1] += NODEWIDTH;

	setMinSize(upleft,upright,downleft);
}

float Forest::getTopY() {
	GLfloat upleft[4], upright[4], downleft[4];
	calcView(upleft,upright,downleft);
	GLfloat tmp[3];
	Matrix_multVect3f(tmp,modelview,upleft,1);
	Vect_copy3f(upleft,tmp);
	Matrix_multVect3f(tmp,modelview,upright,1);
	Vect_copy3f(upright,tmp);
	Matrix_multVect3f(tmp,modelview,downleft,1);
	Vect_copy3f(downleft,tmp);
	getCameraLimits(upleft,upright,downleft,viewport,45,false,1);
	return upleft[1];
}

float Forest::getBottomY() {
	GLfloat upleft[4], upright[4], downleft[4];
	calcView(upleft,upright,downleft);
	GLfloat tmp[3];
	Matrix_multVect3f(tmp,modelview,upleft,1);
	Vect_copy3f(upleft,tmp);
	Matrix_multVect3f(tmp,modelview,upright,1);
	Vect_copy3f(upright,tmp);
	Matrix_multVect3f(tmp,modelview,downleft,1);
	Vect_copy3f(downleft,tmp);
	getCameraLimits(upleft,upright,downleft,viewport,45,false,1);
	return downleft[1];
}

void Forest::getSubTreePos(int tree, Node *n, GLfloat *pos) {
	getTreePos(tree,pos);
	trees[tree]->addPos(n,pos);
	Node **subtree = trees[tree]->getParentLink(n->val);
	n = *subtree;
	if (n->left) pos[0] -= n->left->width;
}

void Forest::getSubTreePos(Tree *t, Node *n, GLfloat *pos) {
	getTreePos(t,pos);
	t->addPos(n,pos);
	Node **subtree = t->getParentLink(n->val);
	n = *subtree;
	if (n->left) pos[0] -= n->left->width;
}

void Forest::getNodePos(int tree, Node *n, GLfloat *pos) {
	getTreePos(tree,pos);
	trees[tree]->addPos(n,pos);
}

void Forest::getNodePos(Tree *t, Node *n, GLfloat *pos) {
	getTreePos(t,pos);
	t->addPos(n,pos);
}


