/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include <math.h>
#include "treeanim.h"
#include "tree.h"
#include "vect.h"
#include "spline.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

enum Type { SCALESPACING, MOVE, GROWPARENT, REMOVEPARENT, FADE };
class TreeAnimItem {
public:
	int starttime;
	int length;
	int type;
	Spline3 size;
	Spline3 move;
	Node *n;
	Tree *t;
};


TreeAnim::TreeAnim(Forest *f) {
	forest = f;
	addValues = NULL;
	remValues = NULL;
	inserted = false;
	allInserted = false;
}

TreeAnim::~TreeAnim() {
	delete forest;
	delete [] addValues;
	delete [] remValues;
}

bool TreeAnim::updateAnim(TreeAnimItem &a) {
	if (a.starttime>animtime) return false;
	int time = animtime - a.starttime;
	float t = float(time)/a.length;
	bool done = false;
	if (t>1) {
		done = true;
		t = 1;
	}
	if (a.type==SCALESPACING) {
		GLfloat size[3];
		a.size.calc(t,size);
		a.n->width = size[0];
		a.n->height = size[1];
	} else if (a.type==MOVE) {
		a.move.calc(t,a.t->pos);
	} else if (a.type==GROWPARENT || a.type==REMOVEPARENT) {
//		float x = t;
		float x = 0.5-0.5*cos(M_PI*t);
		a.n->parentlength = -x;
		if (a.type==REMOVEPARENT) {
			a.n->parentlength = 1 + a.n->parentlength;
			if (done) a.n->parentlength = 0;
		} else if (done) a.n->parentlength = 1;
	} else if (a.type==FADE) {
		a.n->defaultColor = false;
		a.move.calc(t,a.n->color);
		if (done) {
			GLfloat *color = a.n->color;
			GLfloat defaultColor[3];
			forest->getDefaultColor(defaultColor);
			if (fabs(color[0]-defaultColor[0])<0.01 && fabs(color[1]-defaultColor[1])<0.01 && fabs(color[2]-defaultColor[2])<0.01)
				a.n->defaultColor = true;
		}
	}
	return done;
}


bool TreeAnim::updateAnims() {
	int n = anims.size();
	for (int i = 0; i < n; i++) {
		if (updateAnim(anims[i])) {
			anims.erase(anims.begin()+i);
			i--;
			n--;
		}
	}
	return n==0;
}

void TreeAnim::drawAnim(int msec) {
	int starttime = animtime;
	do {
		swapThreads();
	} while (animtime<starttime+msec);
}

void TreeAnim::clearAnims() {
	anims.clear();
}

void TreeAnim::draw(int time, int difftime) {
	forest->draw(time,difftime);
}


void TreeAnim::scaleSpacing(Node *n, float w, float h, bool draw, int offset) {
	TreeAnimItem a;
	a.starttime = animtime+offset;
	a.length = 300;
	a.type = SCALESPACING;
	a.n = n;
	GLfloat startsize[] = { n->width, n->height, 0 }, endsize[] = { w, h, 0 };
	a.size = Spline3(NULL,startsize,endsize,NULL);
	anims.push_back(a);
	if (draw) drawAnim();
}

void TreeAnim::setRelativeToNode(Tree *t, int treenum, Node *n) {
	Tree *desttree = forest->trees[treenum];
	if (desttree->root==NULL) return;
	setAbsolute(t);
	GLfloat rootpos[3];
	forest->getTreePos(treenum,rootpos);
	desttree->addPos(n,rootpos);
	t->type = Tree::TYPE_RELATIVE;
	t->reltree = treenum;
	t->relnode = n;
	Vect_sub3f(t->pos,rootpos);
}

void TreeAnim::setRelativeToRoot(Tree *t, int treenum) {
	Tree *desttree = forest->trees[treenum];
	if (desttree->root==NULL) return;
	setRelativeToNode(t,treenum,desttree->root);
}

void TreeAnim::setAbsolute(Tree *t) {
	if (t->type!=Tree::TYPE_RELATIVE) return;
	GLfloat rootpos[3];
	forest->getTreePos(t->reltree,rootpos);
	forest->trees[t->reltree]->addPos(t->relnode,rootpos);
	Vect_add3f(t->pos,rootpos);
	t->type = Tree::TYPE_ABSOLUTE;
}

void TreeAnim::dropToTree(Tree *t, int treenum, float offsetX, float offsetY) {
	GLfloat pos[3];
	forest->getTreePos(treenum,pos);
	Tree *desttree = forest->trees[treenum];
	if (desttree->root!=NULL)
		desttree->addPos(desttree->root,pos);
	GLfloat startpos[3];
	pos[0] += offsetX;
	Vect_copy3f(startpos,pos);
	startpos[1] = forest->getTopY()+NODEWIDTH;
	pos[1] += offsetY;
	Vect_copy3f(t->pos,startpos);

	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = MOVE;
	a.t = t;
	a.move = Spline3(NULL,startpos,pos,NULL);
	anims.push_back(a);
	drawAnim();
	setRelativeToRoot(t,treenum);
}

void TreeAnim::dropDown(Tree *t, int treenum, float offsetX, float offsetY, bool haspos) {
	GLfloat pos[3];
	if (haspos) {
		Vect_copy3f(pos,t->pos);
	} else {
		forest->getTreePos(treenum,pos);
		Tree *srctree = forest->trees[treenum];
		srctree->root->addPos(t->root,pos);
	}
	pos[0] += offsetX;
	pos[1] += offsetY;
	GLfloat endpos[3];
	Vect_copy3f(endpos,pos);
	endpos[1] = forest->getBottomY()-NODEWIDTH;
	Vect_copy3f(t->pos,pos);

	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = MOVE;
	a.t = t;
	a.move = Spline3(NULL,pos,endpos,NULL);
	anims.push_back(a);
	drawAnim();
}

void TreeAnim::jumpIntoTree(Tree *t, int treenum) {
	setAbsolute(t);
	GLfloat startpos[3];
	Vect_copy3f(startpos,t->pos);
	GLfloat endpos[3], waypoint[3];
	forest->getTreePos(treenum,endpos);
	forest->trees[treenum]->addPos(t->root,endpos);
	Vect_copy3f(waypoint,endpos);
	Vect_add3f(waypoint,startpos);
	Vect_mult3f(waypoint,0.5);
	waypoint[2] += 2*NODEWIDTH;

	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = MOVE;
	a.t = t;
	a.move = Spline3(NULL,startpos,waypoint,endpos);
	anims.push_back(a);
	a.starttime += 300;
	a.move = Spline3(startpos,waypoint,endpos,NULL);
	anims.push_back(a);
	drawAnim();
}

void TreeAnim::growParent(Node *n, bool draw) {
	if (n==NULL) return;
	n->parentlength = 0;
	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.n = n;
	a.type = GROWPARENT;
	anims.push_back(a);
	if (draw) drawAnim();
}

void TreeAnim::removeParent(Node *n, bool draw) {
	if (n==NULL) return;
	n->parentlength = 1;
	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.n = n;
	a.type = REMOVEPARENT;
	anims.push_back(a);
	if (draw) drawAnim();
}

void TreeAnim::fadeColor(Node *n, GLfloat *start, GLfloat *end) {
	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = FADE;
	a.n = n;
	a.move = Spline3(NULL,start,end,NULL);
	anims.push_back(a);
}

void TreeAnim::markColor(Node *n, GLfloat *color) {
	GLfloat startColor[3];
	if (n->defaultColor) forest->getDefaultColor(startColor);
	else memcpy(startColor,n->color,sizeof(GLfloat)*3);
	fadeColor(n,startColor,color);
}

void TreeAnim::markHalfColor(Node *n, GLfloat *color) {
	GLfloat startColor[3];
	if (n->defaultColor) forest->getDefaultColor(startColor);
	else memcpy(startColor,n->color,sizeof(GLfloat)*3);
	GLfloat endColor[3];
	Vect_copy3f(endColor,color);
	Vect_add3f(endColor,startColor);
	Vect_mult3f(endColor,0.5);
	fadeColor(n,startColor,endColor);
}

void TreeAnim::clearMark(Node *n) {
	if (n->defaultColor) return;
	GLfloat startColor[3], endColor[3];
	memcpy(startColor,n->color,sizeof(GLfloat)*3);
	forest->getDefaultColor(endColor);
	fadeColor(n,startColor,endColor);
}

void TreeAnim::setColor(Node *n, GLfloat *color) {
	n->defaultColor = false;
	memcpy(n->color,color,sizeof(GLfloat)*3);
}

void TreeAnim::setHalfColor(Node *n, GLfloat *color) {
	GLfloat startColor[3];
	if (n->defaultColor) forest->getDefaultColor(startColor);
	else memcpy(startColor,n->color,sizeof(GLfloat)*3);
	GLfloat endColor[3];
	Vect_copy3f(endColor,color);
	Vect_add3f(endColor,startColor);
	Vect_mult3f(endColor,0.5);
	memcpy(n->color,endColor,sizeof(GLfloat)*3);
	n->defaultColor = false;
}

void TreeAnim::markRed(Node *n) {
	GLfloat red[] = { 1, 0, 0 };
	markColor(n,red);
	n->red = true;
}

void TreeAnim::setRed(Node *n) {
	GLfloat red[] = { 1, 0, 0 };
	setColor(n,red);
	n->red = true;
}

void TreeAnim::markBlack(Node *n) {
	GLfloat black[] = { 0.15, 0.15, 0.15 };
	markColor(n,black);
	n->red = false;
}

void TreeAnim::setBlack(Node *n) {
	GLfloat black[] = { 0.15, 0.15, 0.15 };
	setColor(n,black);
	n->red = false;
}

void TreeAnim::markGreen(Node *n) {
	GLfloat green[] = { 0, 1, 0 };
	markHalfColor(n,green);
}

void TreeAnim::markGreenRed(Node *n) {
	GLfloat green[] = { 0.5, 0.5, 0 };
	markColor(n,green);
	n->red = true;
}

void TreeAnim::markGreenBlack(Node *n) {
	GLfloat green[] = { 0.15/2, 1.15/2, 0.15/2 };
	markColor(n,green);
	n->red = false;
}

void TreeAnim::setGreen(Node *n) {
	GLfloat green[] = { 0, 1, 0 };
	setHalfColor(n,green);
}

void TreeAnim::moveTree(Tree *t, GLfloat *endpos, bool draw) {
	setAbsolute(t);
	GLfloat startpos[3];
	Vect_copy3f(startpos,t->pos);

	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = MOVE;
	a.t = t;
	a.move = Spline3(NULL,startpos,endpos,NULL);
	anims.push_back(a);
	if (draw) drawAnim();
}

void TreeAnim::jumpTree(Tree *t, GLfloat *endpos, bool draw) {
	setAbsolute(t);
	GLfloat startpos[3];
	Vect_copy3f(startpos,t->pos);
	GLfloat waypoint[3];
	Vect_copy3f(waypoint,startpos);
	Vect_add3f(waypoint,endpos);
	Vect_mult3f(waypoint,0.5);
	waypoint[2] += 2*NODEWIDTH;

	TreeAnimItem a;
	a.starttime = animtime;
	a.length = 300;
	a.type = MOVE;
	a.t = t;
	a.move = Spline3(NULL,startpos,waypoint,endpos);
	anims.push_back(a);
	a.starttime += 300;
	a.move = Spline3(startpos,waypoint,endpos,NULL);
	anims.push_back(a);
	if (draw) drawAnim();
}



int TreeAnim::addTree() {
	Tree *t = new Tree();
	forest->trees.push_back(t);
	if (forest->trees.size()==1) return 0;

	t->root = new Node(-TREESPACING,0);
	scaleSpacing(t->root,0,0);

	delete t->root;
	t->root = NULL;

	return int(forest->trees.size()-1);
}



void TreeAnim::shuffle(int *values, int max) {
	for (int i = 0; i < max; i++) {
		int pos = i+rand()%(max-i);
		int tmp = values[pos];
		values[pos] = values[i];
		values[i] = tmp;
	}
}

void TreeAnim::slightShuffle(int *values, int max) {
	int nswap = max/5 + rand()%(max/5);
	for (int i = 0; i < nswap; i++) {
		int a = rand()%max;
		int diff = rand()%(max/3)-(max/6);
		int b = a+diff;
		if (b<0) b = 0;
		if (b>=max) b = max-1;

		int tmp = values[a];
		values[a] = values[b];
		values[b] = tmp;
	}
	int nshift = max/10 + rand()%(max/10);
	for (int i = 0; i < nshift; i++) {
		int a = rand()%max;
		int b = rand()%max;
		if (b<a) {
			int c = b;
			b = a;
			a = c;
		}

		if (rand()%2) {
			int tmp = values[a];
			for (int i = a; i < b; i++)
				values[i] = values[i+1];
			values[b] = tmp;
		} else {
			int tmp = values[b];
			for (int i = b; i > a; i--)
				values[i] = values[i-1];
			values[a] = tmp;
		}
	}
}

void TreeAnim::initValues(DataOrder order, int n) {
	delete [] addValues;
	delete [] remValues;

	maxval = n;
	addValues = new int[maxval];
	remValues = new int[maxval];
	for (int i = 0; i < maxval; i++) {
		addValues[i] = i+1;
		remValues[i] = i+1;
	}

	if (order==RANDOM) {
		shuffle(addValues,maxval);
		shuffle(remValues,maxval);
//		int values[25] = { 1, 3, 2, 4, 10, 5, 6, 7, 9, 11, 8, 12, 14, 17, 13, 15, 19, 16, 18, 20, 25, 24, 22, 21, 23 };
//		memcpy(addValues,values,sizeof(int)*25);
//		remValues[0] = 10;
	} else if (order==REVERSE) {
		for (int i = 0; i < maxval; i++) {
			addValues[i] = maxval-i;
			remValues[i] = maxval-i;
		}
	} else if (order==SEMISORTED) {
		slightShuffle(addValues,maxval);
		slightShuffle(remValues,maxval);
	}

}

void TreeAnim::copyValues(TreeAnim *ta) {
	delete [] addValues;
	delete [] remValues;

	maxval = ta->maxval;
	addValues = new int[maxval];
	remValues = new int[maxval];
	memcpy(addValues,ta->addValues,sizeof(int)*maxval);
	memcpy(remValues,ta->remValues,sizeof(int)*maxval);
}

void TreeAnim::doneInserting() {
	inserted = true;
	while (!allInserted) swapThreads();
}





