/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <SDL_opengl.h>
#include "vect.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif


void Vect_add3f(GLfloat *v1, const GLfloat *v2) {
	v1[0] += v2[0];
	v1[1] += v2[1];
	v1[2] += v2[2];
}

void Vect_sub3f(GLfloat *v1, const GLfloat *v2) {
	v1[0] -= v2[0];
	v1[1] -= v2[1];
	v1[2] -= v2[2];
}

void Vect_normalize3f(GLfloat *v) {
	float len = Vect_abs3f(v);
	if (len>0) {
		v[0] /= len;
		v[1] /= len;
		v[2] /= len;
	}
}

void Vect_copy3f(GLfloat *out, const GLfloat *in) {
	out[0] = in[0];
	out[1] = in[1];
	out[2] = in[2];
}

void Vect_cross3f(GLfloat *out, const GLfloat *v1, const GLfloat *v2) {
	/*
	 * |  i     j     k  |
	 * |v1[0] v1[1] v1[2]|
	 * |v2[0] v2[1] v2[2]|
	 */
	out[0] = v1[1]*v2[2]-v1[2]*v2[1];
	out[1] = v1[2]*v2[0]-v1[0]*v2[2];
	out[2] = v1[0]*v2[1]-v1[1]*v2[0];
}

float Vect_dot3f(const GLfloat *v1, const GLfloat *v2) {
	return v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2];
}

void Vect_mult3f(GLfloat *v, float f) {
	v[0] *= f;
	v[1] *= f;
	v[2] *= f;
}

float Vect_abs3f(const GLfloat *v) {
	return hypot(hypot(v[0],v[1]),v[2]);
}

Vect3::Vect3() {
}

Vect3::Vect3(const GLfloat *v) {
	for (int i = 0; i < 3; i++)
		vector[i] = v[i];
}

Vect3::Vect3(GLfloat x, GLfloat y, GLfloat z) {
	vector[0] = x;
	vector[1] = y;
	vector[2] = z;
}

Vect3::Vect3(const Vect3 &v) {
	for (int i = 0; i < 3; i++)
		vector[i] = v.vector[i];
}

Vect3 &Vect3::operator+=(const Vect3 &v) {
	for (int i = 0; i < 3; i++)
		vector[i] += v.vector[i];
	return *this;
}

Vect3 Vect3::operator+(const Vect3 &v) const {
	Vect3 out = *this;
	out += v;
	return out;
}

Vect3 &Vect3::operator-=(const Vect3 &v) {
	for (int i = 0; i < 3; i++)
		vector[i] -= v.vector[i];
	return *this;
}

Vect3 Vect3::operator-(const Vect3 &v) const {
	Vect3 out = *this;
	out -= v;
	return out;
}

Vect3 &Vect3::operator*=(GLfloat f) {
	for (int i = 0; i < 3; i++)
		vector[i] *= f;
	return *this;
}

Vect3 Vect3::operator*(GLfloat f) const {
	Vect3 out = *this;
	out *= f;
	return out;
}

Vect3 operator*(GLfloat f, const Vect3 &v) {
	Vect3 out = v;
	out *= f;
	return out;
}

Vect3 &Vect3::operator/=(GLfloat f) {
	for (int i = 0; i < 3; i++)
		vector[i] /= f;
	return *this;
}

Vect3 Vect3::operator/(GLfloat f) const {
	Vect3 out = *this;
	out /= f;
	return out;
}

GLfloat Vect3::dot(const Vect3 &v1, const Vect3 &v2) {
	GLfloat f = 0;
	for (int i = 0; i < 3; i++)
		f += v1.vector[i]*v2.vector[i];
	return f;
}

Vect3 Vect3::cross(const Vect3 &v1, const Vect3 &v2) {
	Vect3 out;
	out.vector[0] = v1.vector[1]*v2.vector[2]-v1.vector[2]*v2.vector[1];
	out.vector[1] = v1.vector[2]*v2.vector[0]-v1.vector[0]*v2.vector[2];
	out.vector[2] = v1.vector[0]*v2.vector[1]-v1.vector[1]*v2.vector[0];
	return out;
}

GLfloat Vect3::abs() const {
	return hypot(hypot(vector[0],vector[1]),vector[2]);
}

Vect3 &Vect3::normalize() {
	float l = abs();
	if (l>0)
		*this /= abs();
	return *this;
}

void Vect3::get(GLfloat *v) const {
	for (int i = 0; i < 3; i++)
		v[i] = vector[i];
}


void printVect3f(GLfloat *vector) {
	printf("[%3.3f %3.3f %3.3f]\n",vector[0],vector[1],vector[2]);
}

void printVect4f(GLfloat *vector) {
	printf("[%3.3f %3.3f %3.3f %3.3f]\n",vector[0],vector[1],vector[2],vector[3]);
}


