/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <vector>
#include <ctype.h>
#include "algorithms.h"

using std::vector;

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#define _WIN32_IE 0x0300
#include <commctrl.h>


class Dialog {
private:
	HINSTANCE hInst;
	HWND window;
	HDC dc;
	const char *classname;
	HFONT font;
	int baseX, baseY;
	int textheight;
public:
	Dialog(const char *classname, LPCSTR icon, WNDPROC wndproc, const char *title) {
		hInst = GetModuleHandle(NULL);
		INITCOMMONCONTROLSEX controls;
		controls.dwSize = sizeof(controls);
		controls.dwICC = ICC_UPDOWN_CLASS;
		InitCommonControlsEx(&controls);

		WNDCLASSEX wc;
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = 0;
		wc.lpfnWndProc = wndproc;
		wc.cbClsExtra = wc.cbWndExtra = 0;
		wc.hInstance = hInst;
		wc.hIcon = LoadIcon(NULL, icon);
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
		wc.lpszMenuName = NULL;
		wc.lpszClassName = classname;
		wc.hIconSm = LoadIcon(NULL, icon);
		RegisterClassEx(&wc);
		this->classname = strdup(classname);

		window = CreateWindowEx(0, classname, title, WS_BORDER | WS_CAPTION | WS_SYSMENU | DS_SETFONT, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInst, NULL);
		DestroyCursor(wc.hCursor);
		DestroyIcon(wc.hIcon);
		DestroyIcon(wc.hIconSm);

		font = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
		SetWindowFont(window,font,FALSE);

		dc = GetDC(window);
		SelectObject(dc,font);
		TEXTMETRIC metric;
		GetTextMetrics(dc,&metric);
		textheight = metric.tmHeight;

		LONG baseunits = GetDialogBaseUnits();
		baseX = LOWORD(baseunits);
		baseY = HIWORD(baseunits);

	}
	~Dialog() {
		DeleteObject(font);
		UnregisterClass(classname,hInst);
		free((void*)classname);
	}
	HWND getWnd() {
		return window;
	}
	void setSize(int w, int h) {
		w = MulDiv(w, baseX, 4);
		h = MulDiv(h, baseY, 8);
		h += GetSystemMetrics(SM_CYCAPTION);
		SetWindowPos(window, HWND_NOTOPMOST, CW_USEDEFAULT, 0, w, h, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
	}
	void dialogUnitsToPixels(int &x, int &y, int &w, int &h) {
		x = MulDiv(x, baseX, 4);
		y = MulDiv(y, baseY, 8);
		w = MulDiv(w, baseX, 4);
		h = MulDiv(h, baseY, 8);
	}
	HWND addControl(const char *cname, const char *title, int x, int y, int w, int h, int styles, int exstyles = 0, int id = -1) {
		dialogUnitsToPixels(x,y,w,h);
		HWND wnd = CreateWindowEx(exstyles, cname, title, WS_CHILD | WS_VISIBLE | styles, x, y, w, h, window, (HMENU)id, hInst, NULL);
		SetWindowFont(wnd,font,FALSE);
		return wnd;
	}
	HWND addGroupBox(const char *title, int x, int y, int w, int h, int styles = 0, int id = -1) {
		return addControl("BUTTON", title, x, y, w, h, styles | BS_GROUPBOX, 0, id);
	}
	HWND addButton(const char *title, int x, int y, int w, int h, bool def = false, int styles = 0, int id = -1) {
		return addControl("BUTTON", title, x, y, w, h, styles | WS_TABSTOP | (def?BS_DEFPUSHBUTTON:BS_PUSHBUTTON), 0, id);
	}
	HWND addCheckBox(const char *title, int x, int y, int w, int h, int styles = 0, int id = -1) {
		return addControl("BUTTON", title, x, y, w, h, styles | WS_TABSTOP | BS_AUTOCHECKBOX, 0, id);
	}
	HWND addRadioButton(const char *title, int x, int y, int w, int h, int styles = 0, int id = -1) {
		return addControl("BUTTON", title, x, y, w, h, styles | WS_TABSTOP | BS_AUTORADIOBUTTON, 0, id);
	}
	HWND addLabel(const char *title, int x, int y, int w, int h, int styles = 0, int id = -1) {
		return addControl("STATIC", title, x, y, w, h, styles | SS_LEFT, 0, id);
	}
	HWND addUpDownEdit(int x, int y, int w, int h, int curval, int min, int max, int id = -1) {
		// ES_READONLY
		HWND edit = addControl("EDIT", "", x, y, w, h, WS_TABSTOP | ES_NUMBER, WS_EX_CLIENTEDGE, id);
		HWND updown = addControl(UPDOWN_CLASS,"", 0, 0, 0, 0, UDS_AUTOBUDDY | UDS_SETBUDDYINT | UDS_ARROWKEYS | UDS_ALIGNRIGHT, 0);
		SendMessage(updown, UDM_SETRANGE, 0, MAKELONG(max, min));
		SendMessage(updown, UDM_SETPOS, 0, MAKELONG(curval, 0));
		return edit;
	}
	HWND addComboBox(int x, int y, int w, int h, int styles = 0, int id = -1) {
		return addControl("COMBOBOX", "", x, y, w, h, styles | WS_TABSTOP | CBS_DROPDOWNLIST, 0, id);
	}

	void run() {
		ReleaseDC(window,dc);
		ShowWindow(window,SW_SHOW);
		UpdateWindow(window);

		MSG msg;
		while (GetMessage(&msg, NULL, 0, 0)>0) {
			if (!IsDialogMessage(window,&msg)) {
				TranslateMessage(&msg);
				if (msg.message==WM_QUIT) break;
				DispatchMessage(&msg);
			}
		}

	}

	int strwidth(const char *str) {
		SIZE size;
		GetTextExtentPoint32(dc,str,strlen(str),&size);
//		return size.cx;
		return MulDiv(size.cx, 4, baseX);
//		RECT rect;
//		rect.left = 0;
//		DrawText(dc,str,strlen(str),&rect,DT_CALCRECT);
	}

};


// TODO: CreateDialogIndirect, DialogBoxIndirect?

static HWND ghWnd;
static HWND ok, cancel, help, fullscreencheck;

struct AlgorithmBox {
	struct algorithm *a;
	Anim::AnimType type;
	HWND box;
};

struct OrderButton {
	DataOrder order;
	HWND button;
};

static vector<AlgorithmBox> boxes;
static vector<OrderButton> buttons;
static bool okclicked = false, destroyed = false;
static HWND lastFocus;

static void showhelp() {
	MessageBox(ghWnd,
		"Available keys:\n"
		"F1 - switch between fullscreen and window\n"
		"left arrow - slow motion\n"
		"right arrow - fast forward\n"
		"space bar - pause\n"
		"escape - exit",
		"Instructions",MB_OK|MB_ICONINFORMATION);
}

static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_CREATE:
		// CreateWindow
		break;
	case WM_DESTROY:
		destroyed = true;
		PostQuitMessage(0);
		break;
	case WM_ACTIVATE:
		if (wParam==WA_INACTIVE)
			lastFocus = GetFocus();
		break;
	case WM_SETFOCUS:
		SetFocus(lastFocus);
		break;
	case WM_COMMAND:
		if (HIWORD(wParam) == BN_CLICKED) {
			bool alg = false;
			bool ord = false;
			HWND control = (HWND)lParam;
			int id = LOWORD(wParam);
			if (id==1) control = ok;
			for (int i = 0; i < int(boxes.size()); i++) {
				if (control==boxes[i].box) {
					alg = true;
					break;
				}
			}
			for (int i = 0; i < int(buttons.size()); i++) {
				if (control==buttons[i].button) {
					ord = true;
					break;
				}
			}

			if (control==ok) {
				if (IsWindowEnabled(ok)) {
					okclicked = true;
					PostQuitMessage(0);
				}
			} else if (control==cancel) {
				PostQuitMessage(0);
			} else if (control==help) {
				showhelp();
			} else if (control==fullscreencheck) {
//				if (Button_GetCheck(control)==BST_CHECKED)
//					Button_SetCheck(control,BST_UNCHECKED);
//				else
//					Button_SetCheck(control,BST_CHECKED);
			} else if (alg) {
//				if (Button_GetCheck(control)==BST_CHECKED)
//					Button_SetCheck(control,BST_UNCHECKED);
//				else
//					Button_SetCheck(control,BST_CHECKED);
				bool checked = false;
				for (int i = 0; i < int(boxes.size()); i++)
					if (Button_GetCheck(boxes[i].box)==BST_CHECKED) checked = true;
				EnableWindow(ok,checked?TRUE:FALSE);
			} else if (ord) {
/*				for (int i = 0; i < int(buttons.size()); i++) {
					if (buttons[i].button==control) {
						Button_SetCheck(buttons[i].button,BST_CHECKED);
					} else {
						Button_SetCheck(buttons[i].button,BST_UNCHECKED);
					}
				}
*/			}
		}
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


void showGui(vector<threadFunc> &funcs, vector<Anim::AnimType> &types, DataOrder &order, vector<const char*> &modes, int &mode, bool &fullscreen, bool nofullscreen, int &sortelements, int &treeelements) {

	Dialog *d = new Dialog("dstgui", IDI_QUESTION, WndProc, "Data structures & Algorithms");

	lastFocus = ghWnd = d->getWnd();



#define GB_TPADDING 10
#define GB_BPADDING 4
#define GB_LPADDING 6
#define GB_RPADDING 4
#define GB_HPADDING (GB_LPADDING+GB_RPADDING)
#define GB_VPADDING (GB_TPADDING+GB_BPADDING)
#define GB_HSPACING 10
#define GB_VSPACING 6

#define W_TPADDING 5
#define W_LPADDING 7
#define W_RPADDING (5+5)
#define W_BPADDING (5+5)

#define CB_H 10
#define UD_H 12
#define UD_SX 30
#define UD_SY 10
#define B_W 40
#define B_H 12
#define COMBO_H 14
#define COMBO_W 40


	int maxwidth1 = 0;
	struct algorithm *a = sortalgorithms;
	int nsort = 0;
	while (a->func!=NULL) {
		nsort++;
		int w = d->strwidth(a->description);
		if (w>maxwidth1) maxwidth1 = w;
		a++;
	}
	a = treealgorithms;
	int ntree = 0;
	int maxwidth2 = 0;
	while (a->func!=NULL) {
		ntree++;
		int w = d->strwidth(a->description);
		if (w>maxwidth2) maxwidth2 = w;
		a++;
	}
	struct OrderDesc *od = orders;
	int norders = 0;
	int maxwidth3 = 35; // COMBO_W
	while (od->name!=NULL) {
		norders++;
		int w = d->strwidth(od->description);
		if (w>maxwidth3) maxwidth3 = w;
		od++;
	}
	maxwidth1 += 12;
	maxwidth2 += 12;
	maxwidth3 += 12;


	int height1 = GB_VPADDING + UD_H + CB_H*nsort;
	int height2 = GB_VPADDING + UD_H + CB_H*ntree;
	int height3 = 2*GB_VPADDING + CB_H*norders + COMBO_H + CB_H + GB_VSPACING;
	int height = height1;
	if (height2>height) height = height2;
	if (height3>height) height = height3;
	int w1 = maxwidth1 + GB_HPADDING + GB_HSPACING;
	int w2 = maxwidth2 + GB_HPADDING + GB_HSPACING;
	int w3 = maxwidth3 + GB_HPADDING + GB_HSPACING;

	int winwidth = W_LPADDING + w1 + w2 + w3 + B_W + W_RPADDING;
	int winheight = W_TPADDING + height + W_BPADDING;
	d->setSize(winwidth,winheight);


	int x = W_LPADDING;
	int y = W_TPADDING;
	int nextx = x + maxwidth1 + GB_HPADDING + GB_HSPACING;
	int nexty = y;
	d->addGroupBox("Sort algorithms", x, y, maxwidth1 + GB_HPADDING, height);

	x += GB_LPADDING;
	y += GB_TPADDING;

	HWND sortedit = d->addUpDownEdit(x, y, UD_SX, UD_SY, sortelements, 5, 10000);
	d->addLabel("elements", x+UD_SX+2, y+UD_SY-8, maxwidth1-(UD_SX+4), 8);
	y += UD_H;

	a = sortalgorithms;
	while (a->func!=NULL) {
		AlgorithmBox ab;
		ab.a = a;
		ab.type = Anim::GRAPH;
		ab.box = d->addCheckBox(a->description, x, y, maxwidth1, CB_H);
		boxes.push_back(ab);
		y += CB_H;
		a++;
	}

	x = nextx;
	y = nexty;
	nextx = x + maxwidth2 + GB_HPADDING + GB_HSPACING;
	nexty = y;
	d->addGroupBox("Tree algorithms", x, y, maxwidth2 + GB_HPADDING, height);

	x += GB_LPADDING;
	y += GB_TPADDING;

	HWND treeedit = d->addUpDownEdit(x, y, UD_SX, UD_SY, treeelements, 5, 10000);
	d->addLabel("elements", x+UD_SX+2, y+UD_SY-8, maxwidth2-(UD_SX+4), 8);
	y += UD_H;

	a = treealgorithms;
	while (a->func!=NULL) {
		AlgorithmBox ab;
		ab.a = a;
		ab.type = Anim::TREE;
		ab.box = d->addCheckBox(a->description, x, y, maxwidth2, CB_H);
		boxes.push_back(ab);
		y += CB_H;
		a++;
	}
	x = nextx;
	y = nexty;

	nextx = x;
	nexty = y + GB_VPADDING + norders*CB_H + GB_VSPACING;
	d->addGroupBox("Data order", x, y, maxwidth3 + GB_HPADDING, GB_VPADDING + norders*CB_H);

	x += GB_LPADDING;
	y += GB_TPADDING;
	od = orders;
	while (od->name!=NULL) {
		OrderButton ob;
		ob.order = od->order;
		ob.button = d->addRadioButton(od->description, x, y, maxwidth3, CB_H, (od==orders?WS_GROUP:0));
		buttons.push_back(ob);
		y += CB_H;
		od++;
	}
	Button_SetCheck(buttons[0].button,BST_CHECKED);
	x = nextx;
	y = nexty;

	nextx = x + maxwidth3+GB_HPADDING + GB_HSPACING;
	nexty = W_TPADDING;
	d->addGroupBox("Display", x, y, maxwidth3 + GB_HPADDING, height - (GB_VPADDING + norders*CB_H + GB_VSPACING));
	x += GB_LPADDING;
	y += GB_TPADDING;
	HWND combo = d->addComboBox(x, y, COMBO_W, COMBO_H + modes.size()*10);
//	Edit_SetReadOnly(GetWindow(combo,GW_CHILD),TRUE);
	int selection = 0;
	for (int i = 0; i < int(modes.size()); i++) {
		if (!strcmp(modes[i],"640x480")) selection = i;
		ComboBox_AddString(combo,modes[i]);
	}
	ComboBox_SetCurSel(combo,selection);
	y += COMBO_H;
	fullscreencheck = d->addCheckBox("Fullscreen", x, y, maxwidth3, CB_H, (nofullscreen?WS_DISABLED:0));
	x = nextx;
	y = nexty;

//	y += GB_TPADDING;
	y += 3;
	ok = d->addButton("OK", x, y, B_W, B_H, true, WS_DISABLED);
	y += B_H + 6;
	cancel = d->addButton("Cancel", x, y, B_W, B_H);
	y += B_H + 6;
	help = d->addButton("Help", x, y, B_W, B_H);


	d->run();

	// GetDlgItem()

	funcs.clear();
	types.clear();
	if (okclicked) {
		for (int i = 0; i < int(boxes.size()); i++) {
			if (Button_GetCheck(boxes[i].box)==BST_CHECKED) {
				funcs.push_back(boxes[i].a->func);
				types.push_back(boxes[i].type);
			}
		}
		for (int i = 0; i < int(buttons.size()); i++) {
			if (Button_GetCheck(buttons[i].button)==BST_CHECKED) {
				order = buttons[i].order;
				break;
			}
		}
		fullscreen = Button_GetCheck(fullscreencheck)==BST_CHECKED;
		mode = ComboBox_GetCurSel(combo);
		char buffer[20];
		Edit_GetText(sortedit,buffer,sizeof(buffer));
		sortelements = atoi(buffer);
		Edit_GetText(treeedit,buffer,sizeof(buffer));
		treeelements = atoi(buffer);
	}

	if (!destroyed) DestroyWindow(ghWnd);
	boxes.clear();
	buttons.clear();

	if (!okclicked) exit(0);

	delete d;

}


